import matplotlib.pyplot as plt
import numpy as np
import pytest


class TestCNNVisualizeFunctions:
    @pytest.fixture
    def X_non_square(self, X_train):
        X = np.hstack(
            (X_train[:, :20 * 28], X_train[:, :20 * 28], X_train[:, :20 * 28]))
        X = X.reshape(-1, 3, 20, 28)
        return X

    def test_plot_loss(self, net_fitted):
        from nolearn.lasagne.visualize import plot_loss
        plot_loss(net_fitted)
        plt.clf()
        plt.cla()

    def plot_conv_weights(self, net, **kwargs):
        from nolearn.lasagne.visualize import plot_conv_weights
        plot_conv_weights(net.layers_['conv1'], **kwargs)
        plt.clf()
        plt.cla()

    @pytest.mark.parametrize('kwargs', [{}, {'figsize': (3, 4)}])
    def test_plot_conv_weights(self, net_fitted, net_color_non_square, kwargs):
        # XXX workaround: fixtures cannot be used (yet) in conjunction
        # with parametrize
        self.plot_conv_weights(net_fitted, **kwargs)
        self.plot_conv_weights(net_color_non_square, **kwargs)

    def plot_conv_activity(self, net, X, **kwargs):
        from nolearn.lasagne.visualize import plot_conv_activity
        plot_conv_activity(net.layers_['conv1'], X, **kwargs)
        plt.clf()
        plt.cla()

    @pytest.mark.parametrize('kwargs', [{}, {'figsize': (3, 4)}])
    def test_plot_conv_activity(
            self, net_fitted, net_color_non_square, X_train, X_non_square,
            kwargs):
        # XXX see above
        self.plot_conv_activity(net_fitted, X_train[:1], **kwargs)
        self.plot_conv_activity(net_fitted, X_train[10:11])

        self.plot_conv_activity(
            net_color_non_square, X_non_square[:1], **kwargs)
        self.plot_conv_activity(
            net_color_non_square, X_non_square[10:11], **kwargs)

    def plot_occlusion(self, net, X, y, **kwargs):
        from nolearn.lasagne.visualize import plot_occlusion
        plot_occlusion(net, X, y, **kwargs)
        plt.clf()
        plt.cla()

    @pytest.mark.parametrize(
        'kwargs', [{}, {'square_length': 3, 'figsize': (3, 4)}])
    def test_plot_occlusion(
            self, net_fitted, net_color_non_square,
            net_with_nonlinearity_layer, X_train, X_non_square, kwargs):
        # XXX see above
        self.plot_occlusion(net_fitted, X_train[3:4], [0], **kwargs)
        self.plot_occlusion(net_fitted, X_train[2:5], [1, 2, 3], **kwargs)

        self.plot_occlusion(
            net_with_nonlinearity_layer, X_train[3:4], [0], **kwargs)
        self.plot_occlusion(
            net_with_nonlinearity_layer, X_train[2:5], [1, 2, 3], **kwargs)

        self.plot_occlusion(
            net_color_non_square, X_non_square[3:4], [0], **kwargs)
        self.plot_occlusion(
            net_color_non_square, X_non_square[2:5], [1, 2, 3], **kwargs)

    def test_draw_to_file_net(self, net_fitted, tmpdir):
        from nolearn.lasagne.visualize import draw_to_file
        fn = str(tmpdir.join('network.pdf'))
        draw_to_file(
            net_fitted, fn, output_shape=False)

    def test_draw_to_notebook_net(self, net_fitted):
        from nolearn.lasagne.visualize import draw_to_notebook
        draw_to_notebook(net_fitted, output_shape=False)

    def test_draw_to_file_layers(self, net_fitted, tmpdir):
        from nolearn.lasagne.visualize import draw_to_file
        fn = str(tmpdir.join('network.pdf'))
        draw_to_file(
            net_fitted.get_all_layers(), fn, output_shape=False)

    def test_draw_to_notebook_layers(self, net_fitted):
        from nolearn.lasagne.visualize import draw_to_notebook
        draw_to_notebook(net_fitted.get_all_layers(), output_shape=False)



import pickle
import sys

from lasagne.layers import get_output
from lasagne.layers import BatchNormLayer
from lasagne.layers import ConcatLayer
from lasagne.layers import Conv2DLayer
from lasagne.layers import DenseLayer
from lasagne.layers import InputLayer
from lasagne.layers import Layer
from lasagne.nonlinearities import identity
from lasagne.nonlinearities import softmax
from lasagne.nonlinearities import sigmoid
from lasagne.objectives import categorical_crossentropy
from lasagne.objectives import aggregate
from lasagne.updates import nesterov_momentum
from mock import Mock
from mock import patch
import numpy as np
import pytest
from sklearn.base import clone
from sklearn.datasets import make_classification
from sklearn.datasets import make_regression
from sklearn.grid_search import GridSearchCV
from sklearn.metrics import accuracy_score
from sklearn.metrics import mean_absolute_error
import theano
import theano.tensor as T

floatX = theano.config.floatX


class TestLayers:
    @pytest.fixture
    def layers(self):
        from nolearn.lasagne.base import Layers
        return Layers([('one', 1), ('two', 2), ('three', 3)])

    def test_getitem_with_key(self, layers):
        assert layers['one'] == 1

    def test_getitem_with_index(self, layers):
        assert layers[0] == 1

    def test_getitem_with_slice(self, layers):
        from nolearn.lasagne.base import Layers
        sliced = layers[:2]
        assert isinstance(sliced, Layers)
        assert sliced.keys() == ['one', 'two']
        assert sliced.values() == [1, 2]

    def test_keys_returns_list(self, layers):
        assert layers.keys() == ['one', 'two', 'three']

    def test_values_returns_list(self, layers):
        assert layers.values() == [1, 2, 3]


class TestFunctionalToy:
    def classif(self, NeuralNet, X, y):
        l = InputLayer(shape=(None, X.shape[1]))
        l = DenseLayer(l, num_units=len(np.unique(y)), nonlinearity=softmax)
        net = NeuralNet(l, update_learning_rate=0.01)
        return net.fit(X, y)

    def classif_no_valid(self, NeuralNet, X, y):
        from nolearn.lasagne import TrainSplit
        l = InputLayer(shape=(None, X.shape[1]))
        l = DenseLayer(l, num_units=len(np.unique(y)), nonlinearity=softmax)
        net = NeuralNet(
            l, update_learning_rate=0.01, train_split=TrainSplit(0))
        return net.fit(X, y)

    def regr(self, NeuralNet, X, y):
        l = InputLayer(shape=(None, X.shape[1]))
        l = DenseLayer(l, num_units=y.shape[1], nonlinearity=None)
        net = NeuralNet(l, regression=True, update_learning_rate=0.01)
        return net.fit(X, y)

    def test_classif_two_classes(self, NeuralNet):
        X, y = make_classification()
        X = X.astype(floatX)
        y = y.astype(np.int32)
        self.classif(NeuralNet, X, y)

    def test_classif_ten_classes(self, NeuralNet):
        X, y = make_classification(n_classes=10, n_informative=10)
        X = X.astype(floatX)
        y = y.astype(np.int32)
        self.classif(NeuralNet, X, y)

    def test_classif_no_valid_two_classes(self, NeuralNet):
        X, y = make_classification()
        X = X.astype(floatX)
        y = y.astype(np.int32)
        self.classif_no_valid(NeuralNet, X, y)

    def test_regr_one_target(self, NeuralNet):
        X, y = make_regression()
        X = X.astype(floatX)
        y = y.reshape(-1, 1).astype(np.float32)
        self.regr(NeuralNet, X, y)

    def test_regr_ten_targets(self, NeuralNet):
        X, y = make_regression(n_targets=10)
        X = X.astype(floatX)
        y = y.astype(floatX)
        self.regr(NeuralNet, X, y)


class TestFunctionalMNIST:
    def test_accuracy(self, net_fitted, mnist, y_pred):
        X, y = mnist
        y_test = y[60000:]
        assert accuracy_score(y_pred, y_test) > 0.85

    def test_train_history(self, net_fitted):
        history = net_fitted.train_history_
        assert len(history) == 2  # due to early stopping
        assert history[1]['valid_accuracy'] > 0.85
        assert history[1]['valid_accuracy'] > history[0]['valid_accuracy']
        assert set(history[0].keys()) == set([
            'dur', 'epoch', 'train_loss', 'train_loss_best',
            'valid_loss', 'valid_loss_best', 'valid_accuracy',
            ])

    def test_early_stopping(self, net_fitted):
        early_stopping = net_fitted.on_epoch_finished[0]
        assert early_stopping.train_history == net_fitted.train_history_

    def test_pickle(self, net_fitted, X_test, y_pred):
        recursionlimit = sys.getrecursionlimit()
        sys.setrecursionlimit(10000)
        pickled = pickle.dumps(net_fitted, -1)
        net_loaded = pickle.loads(pickled)
        assert np.array_equal(net_loaded.predict(X_test), y_pred)
        sys.setrecursionlimit(recursionlimit)

    def test_load_params_from_net(self, net, net_fitted, X_test, y_pred):
        net_loaded = clone(net)
        net_loaded.load_params_from(net_fitted)
        assert np.array_equal(net_loaded.predict(X_test), y_pred)

    def test_load_params_from_params_values(self, net, net_fitted,
                                            X_test, y_pred):
        net_loaded = clone(net)
        net_loaded.load_params_from(net_fitted.get_all_params_values())
        assert np.array_equal(net_loaded.predict(X_test), y_pred)

    def test_save_params_to_path(self, net_fitted, X_test, y_pred):
        path = '/tmp/test_lasagne_functional_mnist.params'
        net_fitted.save_params_to(path)
        net_loaded = clone(net_fitted)
        net_loaded.load_params_from(path)
        assert np.array_equal(net_loaded.predict(X_test), y_pred)

    def test_load_params_from_message(self, net, net_fitted, capsys):
        net2 = clone(net)
        net2.verbose = 1
        net2.load_params_from(net_fitted)

        out = capsys.readouterr()[0]
        message = """\
Loaded parameters to layer 'conv1' (shape 8x1x5x5).
Loaded parameters to layer 'conv1' (shape 8).
Loaded parameters to layer 'conv2' (shape 8x8x5x5).
Loaded parameters to layer 'conv2' (shape 8).
Loaded parameters to layer 'hidden1' (shape 128x128).
Loaded parameters to layer 'hidden1' (shape 128).
Loaded parameters to layer 'output' (shape 128x10).
Loaded parameters to layer 'output' (shape 10).
"""
        assert out == message

    def test_partial_fit(self, net, X_train, y_train):
        net2 = clone(net)
        assert net2.partial_fit(X_train, y_train) is net2
        net2.partial_fit(X_train, y_train)
        history = net2.train_history_
        assert len(history) == 2
        assert history[1]['valid_accuracy'] > 0.85


def test_lasagne_functional_grid_search(mnist, monkeypatch):
    # Make sure that we can satisfy the grid search interface.
    from nolearn.lasagne import NeuralNet

    nn = NeuralNet(
        layers=[],
        )

    param_grid = {
        'more_params': [{'hidden_num_units': 100}, {'hidden_num_units': 200}],
        'update_momentum': [0.9, 0.98],
        }
    X, y = mnist

    vars_hist = []

    def fit(self, X, y):
        vars_hist.append(vars(self).copy())
        return self

    with patch.object(NeuralNet, 'fit', autospec=True) as mock_fit:
        mock_fit.side_effect = fit
        with patch('nolearn.lasagne.NeuralNet.score') as score:
            score.return_value = 0.3
            gs = GridSearchCV(nn, param_grid, cv=2, refit=False, verbose=4)
            gs.fit(X, y)

    assert [entry['update_momentum'] for entry in vars_hist] == [
        0.9, 0.9, 0.98, 0.98] * 2
    assert [entry['more_params'] for entry in vars_hist] == (
        [{'hidden_num_units': 100}] * 4 +
        [{'hidden_num_units': 200}] * 4
        )


def test_clone():
    from nolearn.lasagne import NeuralNet
    from nolearn.lasagne import BatchIterator
    from nolearn.lasagne import objective

    params = dict(
        layers=[
            ('input', InputLayer),
            ('hidden', DenseLayer),
            ('output', DenseLayer),
            ],
        input_shape=(100, 784),
        output_num_units=10,
        output_nonlinearity=softmax,

        more_params={
            'hidden_num_units': 100,
            },
        update=nesterov_momentum,
        update_learning_rate=0.01,
        update_momentum=0.9,

        regression=False,
        objective=objective,
        objective_loss_function=categorical_crossentropy,
        batch_iterator_train=BatchIterator(batch_size=100),
        y_tensor_type=T.ivector,
        use_label_encoder=False,
        on_epoch_finished=None,
        on_training_finished=None,
        max_epochs=100,
        eval_size=0.1,  # BBB
        check_input=True,
        verbose=0,
        )
    nn = NeuralNet(**params)

    nn2 = clone(nn)
    params1 = nn.get_params()
    params2 = nn2.get_params()

    for ignore in (
        'batch_iterator_train',
        'batch_iterator_test',
        'output_nonlinearity',
        'loss',
        'objective',
        'train_split',
        'eval_size',
        'X_tensor_type',
        'on_epoch_finished',
        'on_batch_finished',
        'on_training_started',
        'on_training_finished',
        'custom_scores',
        'scores_train',
        'scores_valid',
            ):
        for par in (params, params1, params2):
            par.pop(ignore, None)

    assert params == params1 == params2


def test_lasagne_functional_regression(boston):
    from nolearn.lasagne import NeuralNet

    X, y = boston

    layer1 = InputLayer(shape=(128, 13))
    layer2 = DenseLayer(layer1, num_units=100)
    output = DenseLayer(layer2, num_units=1, nonlinearity=identity)

    nn = NeuralNet(
        layers=output,
        update_learning_rate=0.01,
        update_momentum=0.1,
        regression=True,
        max_epochs=50,
        )

    nn.fit(X[:300], y[:300])
    assert mean_absolute_error(nn.predict(X[300:]), y[300:]) < 3.0


class TestDefaultObjective:
    @pytest.fixture
    def get_output(self, monkeypatch):
        from nolearn.lasagne import base
        get_output_mock = Mock()
        monkeypatch.setattr(base, 'get_output', get_output_mock)
        return get_output_mock

    @pytest.fixture
    def objective(self):
        from nolearn.lasagne.base import objective
        return objective

    def test_with_defaults(self, objective, get_output):
        loss_function, target = Mock(), Mock()
        loss_function.return_value = np.array([1, 2, 3])
        result = objective(
            [1, 2, 3], loss_function=loss_function, target=target)
        assert result == 2.0
        get_output.assert_called_with(3, deterministic=False)
        loss_function.assert_called_with(get_output.return_value, target)

    def test_with_get_output_kw(self, objective, get_output):
        loss_function, target = Mock(), Mock()
        loss_function.return_value = np.array([1, 2, 3])
        objective(
            [1, 2, 3], loss_function=loss_function, target=target,
            get_output_kw={'i_was': 'here'},
            )
        get_output.assert_called_with(3, deterministic=False, i_was='here')


class TestTrainSplit:
    @pytest.fixture
    def TrainSplit(self):
        from nolearn.lasagne import TrainSplit
        return TrainSplit

    def test_reproducable(self, TrainSplit, nn):
        X, y = np.random.random((100, 10)), np.repeat([0, 1, 2, 3], 25)
        X_train1, X_valid1, y_train1, y_valid1 = TrainSplit(0.2)(
            X, y, nn)
        X_train2, X_valid2, y_train2, y_valid2 = TrainSplit(0.2)(
            X, y, nn)
        assert np.all(X_train1 == X_train2)
        assert np.all(y_valid1 == y_valid2)

    def test_eval_size_zero(self, TrainSplit, nn):
        X, y = np.random.random((100, 10)), np.repeat([0, 1, 2, 3], 25)
        X_train, X_valid, y_train, y_valid = TrainSplit(0.0)(
            X, y, nn)
        assert len(X_train) == len(X)
        assert len(y_train) == len(y)
        assert len(X_valid) == 0
        assert len(y_valid) == 0

    def test_eval_size_half(self, TrainSplit, nn):
        X, y = np.random.random((100, 10)), np.repeat([0, 1, 2, 3], 25)
        X_train, X_valid, y_train, y_valid = TrainSplit(0.51)(
            X, y, nn)
        assert len(X_train) + len(X_valid) == 100
        assert len(y_train) + len(y_valid) == 100
        assert len(X_train) > 45

    def test_regression(self, TrainSplit, nn):
        X = np.random.random((100, 10))
        y = np.random.random((100))
        nn.regression = True
        X_train, X_valid, y_train, y_valid = TrainSplit(0.2)(
            X, y, nn)
        assert len(X_train) == len(y_train) == 80
        assert len(X_valid) == len(y_valid) == 20

    def test_stratified(self, TrainSplit, nn):
        X = np.random.random((100, 10))
        y = np.hstack([np.repeat([0, 0, 0], 25), np.repeat([1], 25)])
        X_train, X_valid, y_train, y_valid = TrainSplit(0.2)(
            X, y, nn)
        assert y_train.sum() == 0.8 * 25
        assert y_valid.sum() == 0.2 * 25

    def test_not_stratified(self, TrainSplit, nn):
        X = np.random.random((100, 10))
        y = np.hstack([np.repeat([0, 0, 0], 25), np.repeat([1], 25)])
        X_train, X_valid, y_train, y_valid = TrainSplit(0.2, stratify=False)(
            X, y, nn)
        assert y_train.sum() == 25
        assert y_valid.sum() == 0

    def test_X_is_dict(self, TrainSplit, nn):
        X = {
            '1': np.random.random((100, 10)),
            '2': np.random.random((100, 10)),
            }
        y = np.repeat([0, 1, 2, 3], 25)

        X_train, X_valid, y_train, y_valid = TrainSplit(0.2)(
            X, y, nn)
        assert len(X_train['1']) == len(X_train['2']) == len(y_train) == 80
        assert len(X_valid['1']) == len(X_valid['2']) == len(y_valid) == 20

    def test_X_is_dict_eval_size_0(self, TrainSplit, nn):
        X = {
            '1': np.random.random((100, 10)),
            '2': np.random.random((100, 10)),
            }
        y = np.repeat([0, 1, 2, 3], 25)

        X_train, X_valid, y_train, y_valid = TrainSplit(0)(
            X, y, nn)
        assert len(X_train['1']) == len(X_train['2']) == len(y_train) == 100
        assert len(X_valid['1']) == len(X_valid['2']) == len(y_valid) == 0


class TestTrainTestSplitBackwardCompatibility:
    @pytest.fixture
    def LegacyNet(self, NeuralNet):
        class LegacyNet(NeuralNet):
            def train_test_split(self, X, y, eval_size):
                self.__call_args__ = (X, y, eval_size)
                split = int(X.shape[0] * eval_size)
                return X[:split], X[split:], y[:split], y[split:]
        return LegacyNet

    def test_legacy_eval_size(self, NeuralNet):
        net = NeuralNet([], eval_size=0.3, max_epochs=0)
        assert net.train_split.eval_size == 0.3

    def test_legacy_method_default_eval_size(self, LegacyNet):
        net = LegacyNet([], max_epochs=0)
        X, y = np.ones((10, 3)), np.zeros(10)
        net.train_loop(X, y)
        assert net.__call_args__ == (X, y, 0.2)

    def test_legacy_method_given_eval_size(self, LegacyNet):
        net = LegacyNet([], eval_size=0.3, max_epochs=0)
        X, y = np.ones((10, 3)), np.zeros(10)
        net.train_loop(X, y)
        assert net.__call_args__ == (X, y, 0.3)


class TestBatchIterator:
    @pytest.fixture
    def BatchIterator(self):
        from nolearn.lasagne import BatchIterator
        return BatchIterator

    @pytest.fixture
    def X(self):
        return np.arange(200).reshape((10, 20)).T.astype('float')

    @pytest.fixture
    def X_dict(self):
        return {
            'one': np.arange(200).reshape((10, 20)).T.astype('float'),
            'two': np.arange(200).reshape((20, 10)).astype('float'),
            }

    @pytest.fixture
    def y(self):
        return np.arange(20)

    @pytest.mark.parametrize("shuffle", [True, False])
    def test_simple_x_and_y(self, BatchIterator, X, y, shuffle):
        bi = BatchIterator(2, shuffle=shuffle)(X, y)
        batches = list(bi)
        assert len(batches) == 10

        X0, y0 = batches[0]
        assert X0.shape == (2, 10)
        assert y0.shape == (2,)

        Xt = np.vstack(b[0] for b in batches)
        yt = np.hstack(b[1] for b in batches)
        assert Xt.shape == X.shape
        assert yt.shape == y.shape
        np.testing.assert_equal(Xt[:, 0], yt)

        if shuffle is False:
            np.testing.assert_equal(X[:2], X0)
            np.testing.assert_equal(y[:2], y0)

    @pytest.mark.parametrize("shuffle", [True, False])
    def test_simple_x_no_y(self, BatchIterator, X, shuffle):
        bi = BatchIterator(2, shuffle=shuffle)(X)
        batches = list(bi)
        assert len(batches) == 10

        X0, y0 = batches[0]
        assert X0.shape == (2, 10)
        assert y0 is None

        if shuffle is False:
            np.testing.assert_equal(X[:2], X0)

    @pytest.mark.parametrize("shuffle", [True, False])
    def test_X_is_dict(self, BatchIterator, X_dict, shuffle):
        bi = BatchIterator(2, shuffle=shuffle)(X_dict)
        batches = list(bi)
        assert len(batches) == 10

        X0, y0 = batches[0]
        assert X0['one'].shape == (2, 10)
        assert X0['two'].shape == (2, 10)
        assert y0 is None

        Xt1 = np.vstack(b[0]['one'] for b in batches)
        Xt2 = np.vstack(b[0]['two'] for b in batches)
        assert Xt1.shape == X_dict['one'].shape
        assert Xt2.shape == X_dict['two'].shape
        np.testing.assert_equal(Xt1[:, 0], Xt2[:, 0] / 10)

        if shuffle is False:
            np.testing.assert_equal(X_dict['one'][:2], X0['one'])
            np.testing.assert_equal(X_dict['two'][:2], X0['two'])

    def test_shuffle_no_copy(self, BatchIterator, X, y):
        bi = BatchIterator(2, shuffle=True)(X, y)
        X0, y0 = list(bi)[0]
        assert X0.base is X  # make sure X0 is a view


class TestCheckForUnusedKwargs:
    def test_okay(self, NeuralNet):
        net = NeuralNet(
            layers=[('input', Mock), ('mylayer', Mock)],
            input_shape=(10, 10),
            mylayer_hey='hey',
            update_foo=1,
            update_bar=2,
            )
        net._create_iter_funcs = lambda *args: (1, 2, 3)
        net.initialize()

    def test_unused(self, NeuralNet):
        net = NeuralNet(
            layers=[('input', Mock), ('mylayer', Mock)],
            input_shape=(10, 10),
            mylayer_hey='hey',
            yourlayer_ho='ho',
            update_foo=1,
            update_bar=2,
            )
        net._create_iter_funcs = lambda *args: (1, 2, 3)

        with pytest.raises(ValueError) as err:
            net.initialize()
        assert str(err.value) == 'Unused kwarg: yourlayer_ho'


class TestInitializeLayers:
    def test_initialization_with_layer_instance(self, NeuralNet):
        layer1 = InputLayer(shape=(128, 13))  # name will be assigned
        layer2 = DenseLayer(layer1, name='output', num_units=2)  # has name
        nn = NeuralNet(layers=layer2)
        out = nn.initialize_layers()
        assert nn.layers_['output'] == layer2 == out
        assert nn.layers_['input0'] == layer1

    def test_initialization_with_layer_instance_bad_params(self, NeuralNet):
        layer = DenseLayer(InputLayer(shape=(128, 13)), num_units=2)
        nn = NeuralNet(layers=layer, dense1_num_units=3)
        with pytest.raises(ValueError):
            nn.initialize_layers()

    def test_initialization_with_tuples(self, NeuralNet):
        input = Mock(__name__='InputLayer', __bases__=(InputLayer,))
        hidden1, hidden2, output = [
            Mock(__name__='MockLayer', __bases__=(Layer,)) for i in range(3)]
        nn = NeuralNet(
            layers=[
                (input, {'shape': (10, 10), 'name': 'input'}),
                (hidden1, {'some': 'param', 'another': 'param'}),
                (hidden2, {}),
                (output, {'name': 'output'}),
                ],
            input_shape=(10, 10),
            mock1_some='iwin',
            )
        out = nn.initialize_layers(nn.layers)

        input.assert_called_with(
            name='input', shape=(10, 10))
        assert nn.layers_['input'] is input.return_value

        hidden1.assert_called_with(
            incoming=input.return_value, name='mock1',
            some='iwin', another='param')
        assert nn.layers_['mock1'] is hidden1.return_value

        hidden2.assert_called_with(
            incoming=hidden1.return_value, name='mock2')
        assert nn.layers_['mock2'] is hidden2.return_value

        output.assert_called_with(
            incoming=hidden2.return_value, name='output')

        assert out is nn.layers_['output']

    def test_initializtion_with_tuples_resolve_layers(self, NeuralNet):
        nn = NeuralNet(
            layers=[
                ('lasagne.layers.InputLayer', {'shape': (None, 10)}),
                ('lasagne.layers.DenseLayer', {'num_units': 33}),
                ],
            )
        out = nn.initialize_layers(nn.layers)
        assert out.num_units == 33

    def test_initialization_legacy(self, NeuralNet):
        input = Mock(__name__='InputLayer', __bases__=(InputLayer,))
        hidden1, hidden2, output = [
            Mock(__name__='MockLayer', __bases__=(Layer,)) for i in range(3)]
        nn = NeuralNet(
            layers=[
                ('input', input),
                ('hidden1', hidden1),
                ('hidden2', hidden2),
                ('output', output),
                ],
            input_shape=(10, 10),
            hidden1_some='param',
            )
        out = nn.initialize_layers(nn.layers)

        input.assert_called_with(
            name='input', shape=(10, 10))
        assert nn.layers_['input'] is input.return_value

        hidden1.assert_called_with(
            incoming=input.return_value, name='hidden1', some='param')
        assert nn.layers_['hidden1'] is hidden1.return_value

        hidden2.assert_called_with(
            incoming=hidden1.return_value, name='hidden2')
        assert nn.layers_['hidden2'] is hidden2.return_value

        output.assert_called_with(
            incoming=hidden2.return_value, name='output')

        assert out is nn.layers_['output']

    def test_initializtion_legacy_resolve_layers(self, NeuralNet):
        nn = NeuralNet(
            layers=[
                ('input', 'lasagne.layers.InputLayer'),
                ('output', 'lasagne.layers.DenseLayer'),
                ],
            input_shape=(None, 10),
            output_num_units=33,
            )
        out = nn.initialize_layers(nn.layers)
        assert out.num_units == 33

    def test_initialization_legacy_with_unicode_names(self, NeuralNet):
        # Test whether legacy initialization is triggered; if not,
        # raises error.
        input = Mock(__name__='InputLayer', __bases__=(InputLayer,))
        hidden1, hidden2, output = [
            Mock(__name__='MockLayer', __bases__=(Layer,)) for i in range(3)]
        nn = NeuralNet(
            layers=[
                (u'input', input),
                (u'hidden1', hidden1),
                (u'hidden2', hidden2),
                (u'output', output),
                ],
            input_shape=(10, 10),
            hidden1_some='param',
            )
        nn.initialize_layers()

    def test_diamond(self, NeuralNet):
        input = Mock(__name__='InputLayer', __bases__=(InputLayer,))
        hidden1, hidden2, concat, output = [
            Mock(__name__='MockLayer', __bases__=(Layer,)) for i in range(4)]
        nn = NeuralNet(
            layers=[
                ('input', input),
                ('hidden1', hidden1),
                ('hidden2', hidden2),
                ('concat', concat),
                ('output', output),
                ],
            input_shape=(10, 10),
            hidden2_incoming='input',
            concat_incomings=['hidden1', 'hidden2'],
            )
        nn.initialize_layers(nn.layers)

        input.assert_called_with(name='input', shape=(10, 10))
        hidden1.assert_called_with(incoming=input.return_value, name='hidden1')
        hidden2.assert_called_with(incoming=input.return_value, name='hidden2')
        concat.assert_called_with(
            incomings=[hidden1.return_value, hidden2.return_value],
            name='concat'
            )
        output.assert_called_with(incoming=concat.return_value, name='output')


class TestCheckGoodInput:
    @pytest.fixture
    def check_good_input(self, nn):
        return nn._check_good_input

    @pytest.fixture
    def X(self):
        return np.arange(100).reshape(10, 10).astype(floatX)

    @pytest.fixture
    def y(self):
        return np.arange(10).astype(np.int32)

    @pytest.fixture
    def y_regr(self):
        return np.arange(10).reshape(-1, 1).astype(floatX)

    def test_X_OK(self, check_good_input, X):
        assert check_good_input(X) == (X, None)

    def test_X_and_y_OK(self, check_good_input, X, y):
        assert check_good_input(X, y) == (X, y)

    def test_X_and_y_OK_regression(self, nn, check_good_input, X, y_regr):
        nn.regression = True
        assert check_good_input(X, y_regr) == (X, y_regr)

    def test_X_and_y_length_mismatch(self, check_good_input, X, y):
        with pytest.raises(ValueError):
            check_good_input(
                X[:9],
                y
                )

    def test_X_dict_and_y_length_mismatch(self, check_good_input, X, y):
        with pytest.raises(ValueError):
            check_good_input(
                {'one': X, 'two': X},
                y[:9],
                )

    def test_X_dict_length_mismatch(self, check_good_input, X):
        with pytest.raises(ValueError):
            check_good_input({
                'one': X,
                'two': X[:9],
                })

    def test_y_regression_1dim(self, nn, check_good_input, X, y_regr):
        y = y_regr.reshape(-1)
        nn.regression = True
        X1, y1 = check_good_input(X, y)
        assert (X1 == X).all()
        assert (y1 == y.reshape(-1, 1)).all()

    def test_y_regression_2dim(self, nn, check_good_input, X, y_regr):
        y = y_regr
        nn.regression = True
        X1, y1 = check_good_input(X, y)
        assert (X1 == X).all()
        assert (y1 == y).all()


class TestGetOutput:
    def test_layer_object(self, net_fitted, X_train):
        layer = net_fitted.layers_['conv2']
        output = net_fitted.get_output(layer, X_train[:3])
        assert output.shape == (3, 8, 8, 8)

    def test_layer_name(self, net_fitted, X_train):
        output = net_fitted.get_output('conv2', X_train[:3])
        assert output.shape == (3, 8, 8, 8)

    def test_get_output_last_layer(self, net_fitted, X_train):
        result = net_fitted.get_output(net_fitted.layers_[-1], X_train[:129])
        expected = net_fitted.predict_proba(X_train[:129])
        np.testing.assert_equal(result, expected)

    def test_no_conv(self, net_no_conv):
        net_no_conv.initialize()
        X = np.random.random((10, 100)).astype(floatX)
        result = net_no_conv.get_output('output', X)
        expected = net_no_conv.predict_proba(X)
        np.testing.assert_equal(result, expected)


class TestMultiInputFunctional:
    @pytest.fixture(scope='session')
    def net(self, NeuralNet):
        return NeuralNet(
            layers=[
                (InputLayer,
                 {'name': 'input1', 'shape': (None, 392)}),
                (DenseLayer,
                 {'name': 'hidden1', 'num_units': 98}),
                (InputLayer,
                 {'name': 'input2', 'shape': (None, 392)}),
                (DenseLayer,
                 {'name': 'hidden2', 'num_units': 98}),
                (ConcatLayer,
                 {'incomings': ['hidden1', 'hidden2']}),
                (DenseLayer,
                 {'name': 'hidden3', 'num_units': 98}),
                (DenseLayer,
                 {'name': 'output', 'num_units': 10, 'nonlinearity': softmax}),
                ],

            update=nesterov_momentum,
            update_learning_rate=0.01,
            update_momentum=0.9,

            max_epochs=2,
            verbose=4,
            )

    @pytest.fixture(scope='session')
    def net_fitted(self, net, mnist):
        X, y = mnist
        X_train, y_train = X[:10000], y[:10000]
        X_train1, X_train2 = X_train[:, :392], X_train[:, 392:]
        return net.fit({'input1': X_train1, 'input2': X_train2}, y_train)

    @pytest.fixture(scope='session')
    def y_pred(self, net_fitted, mnist):
        X, y = mnist
        X_test = X[60000:]
        X_test1, X_test2 = X_test[:, :392], X_test[:, 392:]
        return net_fitted.predict({'input1': X_test1, 'input2': X_test2})

    def test_accuracy(self, net_fitted, mnist, y_pred):
        X, y = mnist
        y_test = y[60000:]
        assert accuracy_score(y_pred, y_test) > 0.85


class TestGradScale:
    @pytest.fixture
    def grad_scale(self):
        from nolearn.lasagne import grad_scale
        return grad_scale

    @pytest.mark.parametrize("layer", [
        BatchNormLayer(InputLayer((None, 16))),
        Conv2DLayer(InputLayer((None, 1, 28, 28)), 2, 3),
        DenseLayer(InputLayer((None, 16)), 16),
        ])
    def test_it(self, grad_scale, layer):
        layer2 = grad_scale(layer, 0.33)
        assert layer2 is layer
        for param in layer.get_params(trainable=True):
            np.testing.assert_almost_equal(param.tag.grad_scale, 0.33)
        for param in layer.get_params(trainable=False):
            assert hasattr(param.tag, 'grad_scale') is False


class TestMultiOutput:
    def test_layers_included(self, NeuralNet):
        def objective(layers_, target, **kwargs):
            out_a_layer = layers_['output_a']
            out_b_layer = layers_['output_b']

            # Get the outputs
            out_a, out_b = get_output([out_a_layer, out_b_layer])

            # Get the targets
            gt_a = T.cast(target[:, 0], 'int32')
            gt_b = target[:, 1].reshape((-1, 1))

            # Calculate the multi task loss
            cls_loss = aggregate(categorical_crossentropy(out_a, gt_a))
            reg_loss = aggregate(categorical_crossentropy(out_b, gt_b))
            loss = cls_loss + reg_loss
            return loss

        # test that both branches of the multi output network are included,
        # and also that a single layer isn't included multiple times.
        l = InputLayer(shape=(None, 1, 28, 28), name="input")
        l = Conv2DLayer(l, name='conv1', filter_size=(5, 5), num_filters=8)
        l = Conv2DLayer(l, name='conv2', filter_size=(5, 5), num_filters=8)

        la = DenseLayer(l, name='hidden_a', num_units=128)
        la = DenseLayer(la, name='output_a', nonlinearity=softmax,
                        num_units=10)

        lb = DenseLayer(l, name='hidden_b', num_units=128)
        lb = DenseLayer(lb, name='output_b', nonlinearity=sigmoid, num_units=1)

        net = NeuralNet(layers=[la, lb],
                        update_learning_rate=0.5,
                        y_tensor_type=None,
                        regression=True,
                        objective=objective)
        net.initialize()

        expected_names = sorted(["input", "conv1", "conv2",
                                 "hidden_a", "output_a",
                                 "hidden_b", "output_b"])
        network_names = sorted(list(net.layers_.keys()))

        assert (expected_names == network_names)



from __future__ import absolute_import

from .._compat import basestring
from .._compat import chain_exception
from .._compat import pickle
from collections import OrderedDict
import itertools
from pydoc import locate
from warnings import warn
from time import time

from lasagne.layers import get_all_layers
from lasagne.layers import get_output
from lasagne.layers import InputLayer
from lasagne.layers import Layer
from lasagne import regularization
from lasagne.objectives import aggregate
from lasagne.objectives import categorical_crossentropy
from lasagne.objectives import squared_error
from lasagne.updates import nesterov_momentum
from lasagne.utils import floatX
from lasagne.utils import unique
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.cross_validation import KFold
from sklearn.cross_validation import StratifiedKFold
from sklearn.metrics import accuracy_score
from sklearn.metrics import mean_squared_error
from sklearn.preprocessing import LabelEncoder
import theano
from theano import tensor as T

from . import PrintLog
from . import PrintLayerInfo


class _list(list):
    pass


class _dict(dict):
    def __contains__(self, key):
        return True


def _sldict(arr, sl):
    if isinstance(arr, dict):
        return {k: v[sl] for k, v in arr.items()}
    else:
        return arr[sl]


class Layers(OrderedDict):
    def __getitem__(self, key):
        if isinstance(key, int):
            return list(self.values()).__getitem__(key)
        elif isinstance(key, slice):
            items = list(self.items()).__getitem__(key)
            return Layers(items)
        else:
            return super(Layers, self).__getitem__(key)

    def keys(self):
        return list(super(Layers, self).keys())

    def values(self):
        return list(super(Layers, self).values())


class BatchIterator(object):
    def __init__(self, batch_size, shuffle=False, seed=42):
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.random = np.random.RandomState(seed)

    def __call__(self, X, y=None):
        if self.shuffle:
            self._shuffle_arrays([X, y] if y is not None else [X], self.random)
        self.X, self.y = X, y
        return self

    def __iter__(self):
        bs = self.batch_size
        for i in range((self.n_samples + bs - 1) // bs):
            sl = slice(i * bs, (i + 1) * bs)
            Xb = _sldict(self.X, sl)
            if self.y is not None:
                yb = self.y[sl]
            else:
                yb = None
            yield self.transform(Xb, yb)

    @classmethod
    def _shuffle_arrays(cls, arrays, random):
        rstate = random.get_state()
        for array in arrays:
            if isinstance(array, dict):
                for v in list(array.values()):
                    random.set_state(rstate)
                    random.shuffle(v)
            else:
                random.set_state(rstate)
                random.shuffle(array)

    @property
    def n_samples(self):
        X = self.X
        if isinstance(X, dict):
            return len(list(X.values())[0])
        else:
            return len(X)

    def transform(self, Xb, yb):
        return Xb, yb

    def __getstate__(self):
        state = dict(self.__dict__)
        for attr in ('X', 'y',):
            if attr in state:
                del state[attr]
        return state


def grad_scale(layer, scale):
    for param in layer.get_params(trainable=True):
        param.tag.grad_scale = floatX(scale)
    return layer


class TrainSplit(object):
    def __init__(self, eval_size, stratify=True):
        self.eval_size = eval_size
        self.stratify = stratify

    def __call__(self, X, y, net):
        if self.eval_size:
            if net.regression or not self.stratify:
                kf = KFold(y.shape[0], round(1. / self.eval_size))
            else:
                kf = StratifiedKFold(y, round(1. / self.eval_size))

            train_indices, valid_indices = next(iter(kf))
            X_train, y_train = _sldict(X, train_indices), y[train_indices]
            X_valid, y_valid = _sldict(X, valid_indices), y[valid_indices]
        else:
            X_train, y_train = X, y
            X_valid, y_valid = _sldict(X, slice(len(y), None)), y[len(y):]

        return X_train, X_valid, y_train, y_valid


class LegacyTrainTestSplit(object):  # BBB
    def __init__(self, eval_size=0.2):
        self.eval_size = eval_size

    def __call__(self, X, y, net):
        return net.train_test_split(X, y, self.eval_size)


def objective(layers,
              loss_function,
              target,
              aggregate=aggregate,
              deterministic=False,
              l1=0,
              l2=0,
              get_output_kw=None):
    if get_output_kw is None:
        get_output_kw = {}
    output_layer = layers[-1]
    network_output = get_output(
        output_layer, deterministic=deterministic, **get_output_kw)
    loss = aggregate(loss_function(network_output, target))

    if l1:
        loss += regularization.regularize_layer_params(
            layers.values(), regularization.l1) * l1
    if l2:
        loss += regularization.regularize_layer_params(
            layers.values(), regularization.l2) * l2
    return loss


class NeuralNet(BaseEstimator):
    """A scikit-learn estimator based on Lasagne.

    """
    def __init__(
        self,
        layers,
        update=nesterov_momentum,
        loss=None,  # BBB
        objective=objective,
        objective_loss_function=None,
        batch_iterator_train=BatchIterator(batch_size=128),
        batch_iterator_test=BatchIterator(batch_size=128),
        regression=False,
        max_epochs=100,
        train_split=TrainSplit(eval_size=0.2),
        custom_scores=None,
        scores_train=None,
        scores_valid=None,
        X_tensor_type=None,
        y_tensor_type=None,
        use_label_encoder=False,
        on_batch_finished=None,
        on_epoch_finished=None,
        on_training_started=None,
        on_training_finished=None,
        more_params=None,
        check_input=True,
        verbose=0,
        **kwargs
        ):
        """:param layers: A list of lasagne layers to compose into the final
                          neural net

        :param on_epoch_finished: A list of functions which are called
                                 after every epoch.  The functions
                                 will be passed the NeuralNet as the
                                 first parameter and its
                                 train_history_ attribute as the
                                 second parameter.

        :param on_training_started: A list of functions which are
                                    called after training has started.
                                    The functions will be passed the
                                    NeuralNet as the first parameter
                                    and its train_history_ attribute
                                    as the second parameter.

        :param on_training_finished: A list of functions which are
                                     called after training is
                                     finished.  The functions will be
                                     passed the NeuralNet as the first
                                     parameter and its train_history_
                                     attribute as the second
                                     parameter.

        """
        if loss is not None:
            raise ValueError(
                "The 'loss' parameter was removed, please use "
                "'objective_loss_function' instead.")  # BBB
        if hasattr(objective, 'get_loss'):
            raise ValueError(
                "The 'Objective' class is no longer supported, please "
                "use 'nolearn.lasagne.objective' or similar.")  # BBB
        if objective_loss_function is None:
            objective_loss_function = (
                squared_error if regression else categorical_crossentropy)

        if hasattr(self, 'train_test_split'):  # BBB
            warn("The 'train_test_split' method has been deprecated, please "
                 "use the 'train_split' parameter instead.")
            train_split = LegacyTrainTestSplit(
                eval_size=kwargs.pop('eval_size', 0.2))

        if 'eval_size' in kwargs:  # BBB
            warn("The 'eval_size' argument has been deprecated, please use "
                 "the 'train_split' parameter instead, e.g.\n"
                 "train_split=TrainSplit(eval_size=0.4)")
            train_split.eval_size = kwargs.pop('eval_size')

        if y_tensor_type is None:
            if regression:
                y_tensor_type = T.TensorType(
                    theano.config.floatX, (False, False))
            else:
                y_tensor_type = T.ivector

        if X_tensor_type is not None:
            raise ValueError(
                "The 'X_tensor_type' parameter has been removed. "
                "It's unnecessary.")  # BBB

        if 'custom_score' in kwargs:
            warn("The 'custom_score' argument has been deprecated, please use "
                 "the 'custom_scores' parameter instead, which is just "
                 "a list of custom scores e.g.\n"
                 "custom_scores=[('first output', lambda y1, y2: abs(y1[0,0]-y2[0,0])), ('second output', lambda y1,y2: abs(y1[0,1]-y2[0,1]))]")

            # add it to custom_scores
            if custom_scores is None:
                custom_scores = [kwargs.pop('custom_score')]
            else:
                custom_scores.append(kwargs.pop('custom_score'))

        if isinstance(layers, Layer):
            layers = _list([layers])
            
        self.layers = layers
        self.update = update
        self.objective = objective
        self.objective_loss_function = objective_loss_function
        self.batch_iterator_train = batch_iterator_train
        self.batch_iterator_test = batch_iterator_test
        self.regression = regression
        self.max_epochs = max_epochs
        self.train_split = train_split
        self.custom_scores = custom_scores
        self.scores_train = scores_train or []
        self.scores_valid = scores_valid or []
        self.y_tensor_type = y_tensor_type
        self.use_label_encoder = use_label_encoder
        self.on_batch_finished = on_batch_finished or []
        self.on_epoch_finished = on_epoch_finished or []
        self.on_training_started = on_training_started or []
        self.on_training_finished = on_training_finished or []
        self.more_params = more_params or {}
        self.check_input = check_input
        self.verbose = verbose
        if self.verbose:
            # XXX: PrintLog should come before any other handlers,
            # because early stopping will otherwise cause the last
            # line not to be printed
            self.on_epoch_finished.append(PrintLog())
            self.on_training_started.append(PrintLayerInfo())

        for key in kwargs.keys():
            assert not hasattr(self, key)
        vars(self).update(kwargs)
        self._kwarg_keys = list(kwargs.keys())

        self.train_history_ = []

        if 'batch_iterator' in kwargs:  # BBB
            raise ValueError(
                "The 'batch_iterator' argument has been replaced. "
                "Use 'batch_iterator_train' and 'batch_iterator_test' instead."
                )

    def _check_for_unused_kwargs(self):
        names = self.layers_.keys() + ['update', 'objective']
        for k in self._kwarg_keys:
            for n in names:
                prefix = '{}_'.format(n)
                if k.startswith(prefix):
                    break
            else:
                raise ValueError("Unused kwarg: {}".format(k))

    def _check_good_input(self, X, y=None):
        if isinstance(X, dict):
            lengths = [len(X1) for X1 in X.values()]
            if len(set(lengths)) > 1:
                raise ValueError("Not all values of X are of equal length.")
            x_len = lengths[0]
        else:
            x_len = len(X)

        if y is not None:
            if len(y) != x_len:
                raise ValueError("X and y are not of equal length.")

        if self.regression and y is not None and y.ndim == 1:
            y = y.reshape(-1, 1)

        return X, y

    def initialize(self):
        if getattr(self, '_initialized', False):
            return

        out = getattr(self, '_output_layer', None)
        if out is None:
            out = self._output_layer = self.initialize_layers()
        self._check_for_unused_kwargs()

        iter_funcs = self._create_iter_funcs(
            self.layers_, self.objective, self.update,
            self.y_tensor_type,
            )
        self.train_iter_, self.eval_iter_, self.predict_iter_ = iter_funcs
        self._initialized = True

    def _get_params_for(self, name):
        collected = {}
        prefix = '{}_'.format(name)

        params = vars(self)
        more_params = self.more_params

        for key, value in itertools.chain(params.items(), more_params.items()):
            if key.startswith(prefix):
                collected[key[len(prefix):]] = value

        return collected

    def _layer_name(self, layer_class, index):
        return "{}{}".format(
            layer_class.__name__.lower().replace("layer", ""), index)

    def initialize_layers(self, layers=None):
        if layers is not None:
            self.layers = layers
        self.layers_ = Layers()

        #If a Layer, or a list of Layers was passed in
        if isinstance(self.layers[0], Layer):
            for out_layer in self.layers:
                for i, layer in enumerate(get_all_layers(out_layer)):
                    if layer not in self.layers_.values():
                        name = layer.name or self._layer_name(layer.__class__, i)
                        self.layers_[name] = layer
                        if self._get_params_for(name) != {}:
                            raise ValueError(
                                "You can't use keyword params when passing a Lasagne "
                                "instance object as the 'layers' parameter of "
                                "'NeuralNet'."
                                )
            return self.layers[-1]

        # 'self.layers' are a list of '(Layer class, kwargs)', so
        # we'll have to actually instantiate the layers given the
        # arguments:
        layer = None
        for i, layer_def in enumerate(self.layers):

            if isinstance(layer_def[1], dict):
                # Newer format: (Layer, {'layer': 'kwargs'})
                layer_factory, layer_kw = layer_def
                layer_kw = layer_kw.copy()
            else:
                # The legacy format: ('name', Layer)
                layer_name, layer_factory = layer_def
                layer_kw = {'name': layer_name}

            if isinstance(layer_factory, str):
                layer_factory = locate(layer_factory)
                assert layer_factory is not None

            if 'name' not in layer_kw:
                layer_kw['name'] = self._layer_name(layer_factory, i)

            more_params = self._get_params_for(layer_kw['name'])
            layer_kw.update(more_params)

            if layer_kw['name'] in self.layers_:
                raise ValueError(
                    "Two layers with name {}.".format(layer_kw['name']))

            # Any layers that aren't subclasses of InputLayer are
            # assumed to require an 'incoming' paramter.  By default,
            # we'll use the previous layer as input:
            if not issubclass(layer_factory, InputLayer):
                if 'incoming' in layer_kw:
                    layer_kw['incoming'] = self.layers_[
                        layer_kw['incoming']]
                elif 'incomings' in layer_kw:
                    layer_kw['incomings'] = [
                        self.layers_[name] for name in layer_kw['incomings']]
                else:
                    layer_kw['incoming'] = layer

            for attr in ('W', 'b'):
                if isinstance(layer_kw.get(attr), str):
                    name = layer_kw[attr]
                    layer_kw[attr] = getattr(self.layers_[name], attr, None)

            try:
                layer_wrapper = layer_kw.pop('layer_wrapper', None)
                layer = layer_factory(**layer_kw)
            except TypeError as e:
                msg = ("Failed to instantiate {} with args {}.\n"
                       "Maybe parameter names have changed?".format(
                           layer_factory, layer_kw))
                chain_exception(TypeError(msg), e)
            self.layers_[layer_kw['name']] = layer
            if layer_wrapper is not None:
                layer = layer_wrapper(layer)
                self.layers_["LW_%s" % layer_kw['name']] = layer

        return layer

    def _create_iter_funcs(self, layers, objective, update, output_type):
        y_batch = output_type('y_batch')

        output_layer = layers[-1]
        objective_kw = self._get_params_for('objective')

        loss_train = objective(
            layers, target=y_batch, **objective_kw)
        loss_eval = objective(
            layers, target=y_batch, deterministic=True, **objective_kw)
        predict_proba = get_output(output_layer, None, deterministic=True)
        if not self.regression:
            predict = predict_proba.argmax(axis=1)
            accuracy = T.mean(T.eq(predict, y_batch))
        else:
            accuracy = loss_eval

        scores_train = [
            s[1](predict_proba, y_batch) for s in self.scores_train]
        scores_valid = [
            s[1](predict_proba, y_batch) for s in self.scores_valid]

        all_params = self.get_all_params(trainable=True)
        grads = theano.grad(loss_train, all_params)
        for idx, param in enumerate(all_params):
            grad_scale = getattr(param.tag, 'grad_scale', 1)
            if grad_scale != 1:
                grads[idx] *= grad_scale
        update_params = self._get_params_for('update')
        updates = update(grads, all_params, **update_params)

        input_layers = [layer for layer in layers.values()
                        if isinstance(layer, InputLayer)]

        X_inputs = [theano.In(input_layer.input_var, name=input_layer.name)
                    for input_layer in input_layers]
        inputs = X_inputs + [theano.In(y_batch, name="y")]

        train_iter = theano.function(
            inputs=inputs,
            outputs=[loss_train] + scores_train,
            updates=updates,
            allow_input_downcast=True,
            )
        eval_iter = theano.function(
            inputs=inputs,
            outputs=[loss_eval, accuracy] + scores_valid,
            allow_input_downcast=True,
            )
        predict_iter = theano.function(
            inputs=X_inputs,
            outputs=predict_proba,
            allow_input_downcast=True,
            )

        return train_iter, eval_iter, predict_iter

    def fit(self, X, y, epochs=None):
        if self.check_input:
            X, y = self._check_good_input(X, y)

        if self.use_label_encoder:
            self.enc_ = LabelEncoder()
            y = self.enc_.fit_transform(y).astype(np.int32)
            self.classes_ = self.enc_.classes_
        self.initialize()

        try:
            self.train_loop(X, y, epochs=epochs)
        except KeyboardInterrupt:
            pass
        return self

    def partial_fit(self, X, y, classes=None):
        return self.fit(X, y, epochs=1)

    def train_loop(self, X, y, epochs=None):
        epochs = epochs or self.max_epochs
        X_train, X_valid, y_train, y_valid = self.train_split(X, y, self)

        on_batch_finished = self.on_batch_finished
        if not isinstance(on_batch_finished, (list, tuple)):
            on_batch_finished = [on_batch_finished]

        on_epoch_finished = self.on_epoch_finished
        if not isinstance(on_epoch_finished, (list, tuple)):
            on_epoch_finished = [on_epoch_finished]

        on_training_started = self.on_training_started
        if not isinstance(on_training_started, (list, tuple)):
            on_training_started = [on_training_started]

        on_training_finished = self.on_training_finished
        if not isinstance(on_training_finished, (list, tuple)):
            on_training_finished = [on_training_finished]

        epoch = 0
        best_valid_loss = (
            min([row['valid_loss'] for row in self.train_history_]) if
            self.train_history_ else np.inf
            )
        best_train_loss = (
            min([row['train_loss'] for row in self.train_history_]) if
            self.train_history_ else np.inf
            )
        for func in on_training_started:
            func(self, self.train_history_)

        num_epochs_past = len(self.train_history_)

        while epoch < epochs:
            epoch += 1

            train_outputs = []
            valid_outputs = []

            if self.custom_scores:
                custom_scores = [[] for _ in self.custom_scores]
            else:
                custom_scores = []

            t0 = time()

            batch_train_sizes = []
            for Xb, yb in self.batch_iterator_train(X_train, y_train):
                train_outputs.append(
                    self.apply_batch_func(self.train_iter_, Xb, yb))
                batch_train_sizes.append(len(Xb))

                for func in on_batch_finished:
                    func(self, self.train_history_)

            batch_valid_sizes = []
            for Xb, yb in self.batch_iterator_test(X_valid, y_valid):
                valid_outputs.append(
                    self.apply_batch_func(self.eval_iter_, Xb, yb))
                batch_valid_sizes.append(len(Xb))

                if self.custom_scores:
                    y_prob = self.apply_batch_func(self.predict_iter_, Xb)
                    for custom_scorer, custom_score in zip(
                            self.custom_scores, custom_scores):
                        custom_score.append(custom_scorer[1](yb, y_prob))

            train_outputs = np.array(train_outputs, dtype=object).T
            train_outputs = [
                np.average(
                    [np.mean(row) for row in col],
                    weights=batch_train_sizes,
                    )
                for col in train_outputs
                ]

            if valid_outputs:
                valid_outputs = np.array(valid_outputs, dtype=object).T
                valid_outputs = [
                    np.average(
                        [np.mean(row) for row in col],
                        weights=batch_valid_sizes,
                        )
                    for col in valid_outputs
                    ]

            if custom_scores:
                avg_custom_scores = np.average(
                    custom_scores, weights=batch_valid_sizes, axis=1)

            if train_outputs[0] < best_train_loss:
                best_train_loss = train_outputs[0]
            if valid_outputs and valid_outputs[0] < best_valid_loss:
                best_valid_loss = valid_outputs[0]

            info = {
                'epoch': num_epochs_past + epoch,
                'train_loss': train_outputs[0],
                'train_loss_best': best_train_loss == train_outputs[0],
                'valid_loss': valid_outputs[0]
                if valid_outputs else np.nan,
                'valid_loss_best': best_valid_loss == valid_outputs[0]
                if valid_outputs else np.nan,
                'valid_accuracy': valid_outputs[1]
                if valid_outputs else np.nan,
                'dur': time() - t0,
                }

            if self.custom_scores:
                for index, custom_score in enumerate(self.custom_scores):
                    info[custom_score[0]] = avg_custom_scores[index]

            if self.scores_train:
                for index, (name, func) in enumerate(self.scores_train):
                    info[name] = train_outputs[index + 1]

            if self.scores_valid:
                for index, (name, func) in enumerate(self.scores_valid):
                    info[name] = valid_outputs[index + 2]

            self.train_history_.append(info)

            try:
                for func in on_epoch_finished:
                    func(self, self.train_history_)
            except StopIteration:
                break

        for func in on_training_finished:
            func(self, self.train_history_)

    @staticmethod
    def apply_batch_func(func, Xb, yb=None):
        if isinstance(Xb, dict):
            kwargs = dict(Xb)
            if yb is not None:
                kwargs['y'] = yb
            return func(**kwargs)
        else:
            return func(Xb) if yb is None else func(Xb, yb)

    def predict_proba(self, X):
        probas = []
        for Xb, yb in self.batch_iterator_test(X):
            probas.append(self.apply_batch_func(self.predict_iter_, Xb))
        return np.vstack(probas)

    def predict(self, X):
        if self.regression:
            return self.predict_proba(X)
        else:
            y_pred = np.argmax(self.predict_proba(X), axis=1)
            if self.use_label_encoder:
                y_pred = self.enc_.inverse_transform(y_pred)
            return y_pred

    def get_output(self, layer, X):
        if isinstance(layer, basestring):
            layer = self.layers_[layer]

        fn_cache = getattr(self, '_get_output_fn_cache', None)
        if fn_cache is None:
            fn_cache = {}
            self._get_output_fn_cache = fn_cache

        if layer not in fn_cache:
            xs = self.layers_[0].input_var.type()
            get_activity = theano.function([xs], get_output(layer, xs))
            fn_cache[layer] = get_activity
        else:
            get_activity = fn_cache[layer]

        outputs = []
        for Xb, yb in self.batch_iterator_test(X):
            outputs.append(get_activity(Xb))
        return np.vstack(outputs)

    def score(self, X, y):
        score = mean_squared_error if self.regression else accuracy_score
        return float(score(self.predict(X), y))

    def get_all_layers(self):
        return self.layers_.values()

    def get_all_params(self, **kwargs):
        layers = self.get_all_layers()
        params = sum([l.get_params(**kwargs) for l in layers], [])
        return unique(params)

    def get_all_params_values(self):
        return_value = OrderedDict()
        for name, layer in self.layers_.items():
            return_value[name] = [p.get_value() for p in layer.get_params()]
        return return_value

    def load_params_from(self, source):
        self.initialize()

        if isinstance(source, basestring):
            with open(source, 'rb') as f:
                source = pickle.load(f)

        if isinstance(source, NeuralNet):
            source = source.get_all_params_values()

        success = "Loaded parameters to layer '{}' (shape {})."
        failure = ("Could not load parameters to layer '{}' because "
                   "shapes did not match: {} vs {}.")

        for key, values in source.items():
            layer = self.layers_.get(key)
            if layer is not None:
                for p1, p2v in zip(layer.get_params(), values):
                    shape1 = p1.get_value().shape
                    shape2 = p2v.shape
                    shape1s = 'x'.join(map(str, shape1))
                    shape2s = 'x'.join(map(str, shape2))
                    if shape1 == shape2:
                        p1.set_value(p2v)
                        if self.verbose:
                            print(success.format(
                                key, shape1s, shape2s))
                    else:
                        if self.verbose:
                            print(failure.format(
                                key, shape1s, shape2s))

    def save_params_to(self, fname):
        params = self.get_all_params_values()
        with open(fname, 'wb') as f:
            pickle.dump(params, f, -1)

    def load_weights_from(self, source):
        warn("The 'load_weights_from' method will be removed in nolearn 0.6. "
             "Please use 'load_params_from' instead.")

        if isinstance(source, list):
            raise ValueError(
                "Loading weights from a list of parameter values is no "
                "longer supported.  Please send me something like the "
                "return value of 'net.get_all_params_values()' instead.")

        return self.load_params_from(source)

    def save_weights_to(self, fname):
        warn("The 'save_weights_to' method will be removed in nolearn 0.6. "
             "Please use 'save_params_to' instead.")
        return self.save_params_to(fname)

    def __setstate__(self, state):  # BBB for pickles that don't have the graph
        self.__dict__.update(state)
        self.initialize()

    def get_params(self, deep=True):
        params = super(NeuralNet, self).get_params(deep=deep)

        # Incidentally, Lasagne layers have a 'get_params' too, which
        # for sklearn's 'clone' means it would treat it in a special
        # way when cloning.  Wrapping the list of layers in a custom
        # list type does the trick here, but of course it's crazy:
        params['layers'] = _list(params['layers'])
        return _dict(params)

    def _get_param_names(self):
        # This allows us to have **kwargs in __init__ (woot!):
        param_names = super(NeuralNet, self)._get_param_names()
        return param_names + self._kwarg_keys



from itertools import product

from lasagne.layers import get_output
from lasagne.layers import get_output_shape
from lasagne.objectives import binary_crossentropy
import matplotlib.pyplot as plt
import numpy as np
import theano
import theano.tensor as T
import io
import lasagne



def plot_loss(net):
    train_loss = [row['train_loss'] for row in net.train_history_]
    valid_loss = [row['valid_loss'] for row in net.train_history_]
    plt.plot(train_loss, label='train loss')
    plt.plot(valid_loss, label='valid loss')
    plt.xlabel('epoch')
    plt.ylabel('loss')
    plt.legend(loc='best')
    return plt


def plot_conv_weights(layer, figsize=(6, 6)):
    """Plot the weights of a specific layer.

    Only really makes sense with convolutional layers.

    Parameters
    ----------
    layer : lasagne.layers.Layer

    """
    W = layer.W.get_value()
    shape = W.shape
    nrows = np.ceil(np.sqrt(shape[0])).astype(int)
    ncols = nrows

    for feature_map in range(shape[1]):
        figs, axes = plt.subplots(nrows, ncols, figsize=figsize, squeeze=False)

        for ax in axes.flatten():
            ax.set_xticks([])
            ax.set_yticks([])
            ax.axis('off')

        for i, (r, c) in enumerate(product(range(nrows), range(ncols))):
            if i >= shape[0]:
                break
            axes[r, c].imshow(W[i, feature_map], cmap='gray',
                              interpolation='none')
    return plt


def plot_conv_activity(layer, x, figsize=(6, 8)):
    """Plot the acitivities of a specific layer.

    Only really makes sense with layers that work 2D data (2D
    convolutional layers, 2D pooling layers ...).

    Parameters
    ----------
    layer : lasagne.layers.Layer

    x : numpy.ndarray
      Only takes one sample at a time, i.e. x.shape[0] == 1.

    """
    if x.shape[0] != 1:
        raise ValueError("Only one sample can be plotted at a time.")

    # compile theano function
    xs = T.tensor4('xs').astype(theano.config.floatX)
    get_activity = theano.function([xs], get_output(layer, xs))

    activity = get_activity(x)
    shape = activity.shape
    nrows = np.ceil(np.sqrt(shape[1])).astype(int)
    ncols = nrows

    figs, axes = plt.subplots(nrows + 1, ncols, figsize=figsize, squeeze=False)
    axes[0, ncols // 2].imshow(1 - x[0][0], cmap='gray',
                               interpolation='none')
    axes[0, ncols // 2].set_title('original')

    for ax in axes.flatten():
        ax.set_xticks([])
        ax.set_yticks([])
        ax.axis('off')

    for i, (r, c) in enumerate(product(range(nrows), range(ncols))):
        if i >= shape[1]:
            break
        ndim = activity[0][i].ndim
        if ndim != 2:
            raise ValueError("Wrong number of dimensions, image data should "
                             "have 2, instead got {}".format(ndim))
        axes[r + 1, c].imshow(-activity[0][i], cmap='gray',
                              interpolation='none')
    return plt


def occlusion_heatmap(net, x, target, square_length=7):
    """An occlusion test that checks an image for its critical parts.

    In this function, a square part of the image is occluded (i.e. set
    to 0) and then the net is tested for its propensity to predict the
    correct label. One should expect that this propensity shrinks of
    critical parts of the image are occluded. If not, this indicates
    overfitting.

    Depending on the depth of the net and the size of the image, this
    function may take awhile to finish, since one prediction for each
    pixel of the image is made.

    Currently, all color channels are occluded at the same time. Also,
    this does not really work if images are randomly distorted by the
    batch iterator.

    See paper: Zeiler, Fergus 2013

    Parameters
    ----------
    net : NeuralNet instance
      The neural net to test.

    x : np.array
      The input data, should be of shape (1, c, x, y). Only makes
      sense with image data.

    target : int
      The true value of the image. If the net makes several
      predictions, say 10 classes, this indicates which one to look
      at.

    square_length : int (default=7)
      The length of the side of the square that occludes the image.
      Must be an odd number.

    Results
    -------
    heat_array : np.array (with same size as image)
      An 2D np.array that at each point (i, j) contains the predicted
      probability of the correct class if the image is occluded by a
      square with center (i, j).

    """
    if (x.ndim != 4) or x.shape[0] != 1:
        raise ValueError("This function requires the input data to be of "
                         "shape (1, c, x, y), instead got {}".format(x.shape))
    if square_length % 2 == 0:
        raise ValueError("Square length has to be an odd number, instead "
                         "got {}.".format(square_length))

    num_classes = get_output_shape(net.layers_[-1])[1]
    img = x[0].copy()
    bs, col, s0, s1 = x.shape

    heat_array = np.zeros((s0, s1))
    pad = square_length // 2 + 1
    x_occluded = np.zeros((s1, col, s0, s1), dtype=img.dtype)
    probs = np.zeros((s0, s1, num_classes))

    # generate occluded images
    for i in range(s0):
        # batch s1 occluded images for faster prediction
        for j in range(s1):
            x_pad = np.pad(img, ((0, 0), (pad, pad), (pad, pad)), 'constant')
            x_pad[:, i:i + square_length, j:j + square_length] = 0.
            x_occluded[j] = x_pad[:, pad:-pad, pad:-pad]
        y_proba = net.predict_proba(x_occluded)
        probs[i] = y_proba.reshape(s1, num_classes)

    # from predicted probabilities, pick only those of target class
    for i in range(s0):
        for j in range(s1):
            heat_array[i, j] = probs[i, j, target]
    return heat_array


def _plot_heat_map(net, X, figsize, get_heat_image):
    if (X.ndim != 4):
        raise ValueError("This function requires the input data to be of "
                         "shape (b, c, x, y), instead got {}".format(X.shape))

    num_images = X.shape[0]
    if figsize[1] is None:
        figsize = (figsize[0], num_images * figsize[0] / 3)
    figs, axes = plt.subplots(num_images, 3, figsize=figsize)

    for ax in axes.flatten():
        ax.set_xticks([])
        ax.set_yticks([])
        ax.axis('off')

    for n in range(num_images):
        heat_img = get_heat_image(net, X[n:n + 1, :, :, :], n)

        ax = axes if num_images == 1 else axes[n]
        img = X[n, :, :, :].mean(0)
        ax[0].imshow(-img, interpolation='nearest', cmap='gray')
        ax[0].set_title('image')
        ax[1].imshow(-heat_img, interpolation='nearest', cmap='Reds')
        ax[1].set_title('critical parts')
        ax[2].imshow(-img, interpolation='nearest', cmap='gray')
        ax[2].imshow(-heat_img, interpolation='nearest', cmap='Reds',
                     alpha=0.6)
        ax[2].set_title('super-imposed')
    return plt


def plot_occlusion(net, X, target, square_length=7, figsize=(9, None)):
    """Plot which parts of an image are particularly import for the
    net to classify the image correctly.

    See paper: Zeiler, Fergus 2013

    Parameters
    ----------
    net : NeuralNet instance
      The neural net to test.

    X : numpy.array
      The input data, should be of shape (b, c, 0, 1). Only makes
      sense with image data.

    target : list or numpy.array of ints
      The true values of the image. If the net makes several
      predictions, say 10 classes, this indicates which one to look
      at. If more than one sample is passed to X, each of them needs
      its own target.

    square_length : int (default=7)
      The length of the side of the square that occludes the image.
      Must be an odd number.

    figsize : tuple (int, int)
      Size of the figure.

    Plots
    -----
    Figure with 3 subplots: the original image, the occlusion heatmap,
    and both images super-imposed.

    """
    return _plot_heat_map(
        net, X, figsize, lambda net, X, n: occlusion_heatmap(
            net, X, target[n], square_length))


def saliency_map(input, output, pred, X):
    score = -binary_crossentropy(output[:, pred], np.array([1])).sum()
    return np.abs(T.grad(score, input).eval({input: X}))


def saliency_map_net(net, X):
    input = net.layers_[0].input_var
    output = get_output(net.layers_[-1])
    pred = output.eval({input: X}).argmax(axis=1)
    return saliency_map(input, output, pred, X)[0].transpose(1, 2, 0).squeeze()


def plot_saliency(net, X, figsize=(9, None)):
    return _plot_heat_map(
        net, X, figsize, lambda net, X, n: -saliency_map_net(net, X))


def get_hex_color(layer_type):
    """
    Determines the hex color for a layer.
    :parameters:
        - layer_type : string
            Class name of the layer
    :returns:
        - color : string containing a hex color for filling block.
    """
    COLORS = ['#4A88B3', '#98C1DE', '#6CA2C8', '#3173A2', '#17649B',
              '#FFBB60', '#FFDAA9', '#FFC981', '#FCAC41', '#F29416',
              '#C54AAA', '#E698D4', '#D56CBE', '#B72F99', '#B0108D',
              '#75DF54', '#B3F1A0', '#91E875', '#5DD637', '#3FCD12']

    hashed = int(hash(layer_type)) % 5

    if "conv" in layer_type.lower():
        return COLORS[:5][hashed]
    if layer_type in lasagne.layers.pool.__all__:
        return COLORS[5:10][hashed]
    if layer_type in lasagne.layers.recurrent.__all__:
        return COLORS[10:15][hashed]
    else:
        return COLORS[15:20][hashed]


def make_pydot_graph(layers, output_shape=True, verbose=False):
    """
    :parameters:
        - layers : list
            List of the layers, as obtained from lasagne.layers.get_all_layers
        - output_shape: (default `True`)
            If `True`, the output shape of each layer will be displayed.
        - verbose: (default `False`)
            If `True`, layer attributes like filter shape, stride, etc.
            will be displayed.
    :returns:
        - pydot_graph : PyDot object containing the graph
    """
    import pydotplus as pydot
    pydot_graph = pydot.Dot('Network', graph_type='digraph')
    pydot_nodes = {}
    pydot_edges = []
    for i, layer in enumerate(layers):
        layer_type = '{0}'.format(layer.__class__.__name__)
        key = repr(layer)
        label = layer_type
        color = get_hex_color(layer_type)
        if verbose:
            for attr in ['num_filters', 'num_units', 'ds',
                         'filter_shape', 'stride', 'strides', 'p']:
                if hasattr(layer, attr):
                    label += '\n{0}: {1}'.format(attr, getattr(layer, attr))
            if hasattr(layer, 'nonlinearity'):
                try:
                    nonlinearity = layer.nonlinearity.__name__
                except AttributeError:
                    nonlinearity = layer.nonlinearity.__class__.__name__
                label += '\nnonlinearity: {0}'.format(nonlinearity)

        if output_shape:
            label += '\nOutput shape: {0}'.format(layer.output_shape)

        pydot_nodes[key] = pydot.Node(
            key, label=label, shape='record', fillcolor=color, style='filled')

        if hasattr(layer, 'input_layers'):
            for input_layer in layer.input_layers:
                pydot_edges.append([repr(input_layer), key])

        if hasattr(layer, 'input_layer'):
            pydot_edges.append([repr(layer.input_layer), key])

    for node in pydot_nodes.values():
        pydot_graph.add_node(node)

    for edges in pydot_edges:
        pydot_graph.add_edge(
            pydot.Edge(pydot_nodes[edges[0]], pydot_nodes[edges[1]]))
    return pydot_graph


def draw_to_file(layers, filename, **kwargs):
    """
    Draws a network diagram to a file
    :parameters:
        - layers : list or NeuralNet instance
            List of layers or the neural net to draw.
        - filename : string
            The filename to save output to
        - **kwargs: see docstring of make_pydot_graph for other options
    """
    layers = (layers.get_all_layers() if hasattr(layers, 'get_all_layers')
              else layers)
    dot = make_pydot_graph(layers, **kwargs)
    ext = filename[filename.rfind('.') + 1:]
    with io.open(filename, 'wb') as fid:
        fid.write(dot.create(format=ext))


def draw_to_notebook(layers, **kwargs):
    """
    Draws a network diagram in an IPython notebook
    :parameters:
        - layers : list or NeuralNet instance
            List of layers or the neural net to draw.
        - **kwargs : see the docstring of make_pydot_graph for other options
    """
    from IPython.display import Image
    layers = (layers.get_all_layers() if hasattr(layers, 'get_all_layers')
              else layers)
    dot = make_pydot_graph(layers, **kwargs)
    return Image(dot.create_png())



from .handlers import (
    PrintLayerInfo,
    PrintLog,
    RememberBestWeights,
    SaveWeights,
    WeightLog,
    )
from .base import (
    BatchIterator,
    grad_scale,
    objective,
    NeuralNet,
    TrainSplit,
    )



from datetime import timedelta
from time import time
import warnings

from gdbn.dbn import buildDBN
from gdbn import activationFunctions
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.preprocessing import LabelEncoder
from sklearn.preprocessing import OneHotEncoder

warnings.warn("""\
The nolearn.dbn module is no longer supported.  Take a look at
*nolearn.lasagne* for a more modern neural net toolkit.
""")


class DBN(BaseEstimator):  # pragma: no cover
    """A scikit-learn estimator based on George Dahl's DBN
    implementation `gdbn`.
    """
    def __init__(
        self,
        layer_sizes=None,
        scales=0.05,
        fan_outs=None,
        output_act_funct=None,
        real_valued_vis=True,
        use_re_lu=True,
        uniforms=False,

        learn_rates=0.1,
        learn_rate_decays=1.0,
        learn_rate_minimums=0.0,
        momentum=0.9,
        l2_costs=0.0001,
        dropouts=0,
        nesterov=True,
        nest_compare=True,
        rms_lims=None,

        learn_rates_pretrain=None,
        momentum_pretrain=None,
        l2_costs_pretrain=None,
        nest_compare_pretrain=None,

        epochs=10,
        epochs_pretrain=0,
        loss_funct=None,
        minibatch_size=64,
        minibatches_per_epoch=None,

        pretrain_callback=None,
        fine_tune_callback=None,

        random_state=None,

        verbose=0,
        ):
        """
        Many parameters such as `learn_rates`, `dropouts` etc. will
        also accept a single value, in which case that value will be
        used for all layers.  To control the value per layer, pass a
        list of values instead; see examples below.

        Parameters ending with `_pretrain` may be provided to override
        the given parameter for pretraining.  Consider an example
        where you want the pre-training to use a lower learning rate
        than the fine tuning (the backprop), then you'd maybe pass
        something like::

          DBN([783, 300, 10], learn_rates=0.1, learn_rates_pretrain=0.005)

        If you don't pass the `learn_rates_pretrain` parameter, the
        value of `learn_rates` will be used for both pre-training and
        fine tuning.  (Which seems to not work very well.)

        :param layer_sizes: A list of integers of the form
                            ``[n_vis_units, n_hid_units1,
                            n_hid_units2, ..., n_out_units]``.

                            An example: ``[784, 300, 10]``

                            The number of units in the input layer and
                            the output layer will be set automatically
                            if you set them to -1.  Thus, the above
                            example is equivalent to ``[-1, 300, -1]``
                            if you pass an ``X`` with 784 features,
                            and a ``y`` with 10 classes.

        :param scales: Scale of the randomly initialized weights.  A
                       list of floating point values.  When you find
                       good values for the scale of the weights you
                       can speed up training a lot, and also improve
                       performance.  Defaults to `0.05`.

        :param fan_outs: Number of nonzero incoming connections to a
                         hidden unit.  Defaults to `None`, which means
                         that all connections have non-zero weights.

        :param output_act_funct: Output activation function.  Instance
                                 of type
                                 :class:`~gdbn.activationFunctions.Sigmoid`,
                                 :class:`~.gdbn.activationFunctions.Linear`,
                                 :class:`~.gdbn.activationFunctions.Softmax`
                                 from the
                                 :mod:`gdbn.activationFunctions`
                                 module.  Defaults to
                                 :class:`~.gdbn.activationFunctions.Softmax`.

        :param real_valued_vis: Set `True` (the default) if visible
                                units are real-valued.

        :param use_re_lu: Set `True` to use rectified linear units.
                          Defaults to `False`.

        :param uniforms: Not documented at this time.

        :param learn_rates: A list of learning rates, one entry per
                            weight layer.

                            An example: ``[0.1, 0.1]``

        :param learn_rate_decays: The number with which the
                                  `learn_rate` is multiplied after
                                  each epoch of fine-tuning.

        :param learn_rate_minimums: The minimum `learn_rates`; after
                                    the learn rate reaches the minimum
                                    learn rate, the
                                    `learn_rate_decays` no longer has
                                    any effect.

        :param momentum: Momentum

        :param l2_costs: L2 costs per weight layer.

        :param dropouts: Dropouts per weight layer.

        :param nesterov: Not documented at this time.

        :param nest_compare: Not documented at this time.

        :param rms_lims: Not documented at this time.

        :param learn_rates_pretrain: A list of learning rates similar
                                     to `learn_rates_pretrain`, but
                                     used for pretraining.  Defaults
                                     to value of `learn_rates` parameter.

        :param momentum_pretrain: Momentum for pre-training.  Defaults
                                  to value of `momentum` parameter.

        :param l2_costs_pretrain: L2 costs per weight layer, for
                                  pre-training.  Defaults to the value
                                  of `l2_costs` parameter.

        :param nest_compare_pretrain: Not documented at this time.

        :param epochs: Number of epochs to train (with backprop).

        :param epochs_pretrain: Number of epochs to pre-train (with CDN).

        :param loss_funct: A function that calculates the loss.  Used
                           for displaying learning progress and for
                           :meth:`score`.

        :param minibatch_size: Size of a minibatch.

        :param minibatches_per_epoch: Number of minibatches per epoch.
                                      The default is to use as many as
                                      fit into our training set.

        :param pretrain_callback: An optional function that takes as
                                  arguments the :class:`DBN` instance,
                                  the epoch and the layer index as its
                                  argument, and is called for each
                                  epoch of pretraining.

        :param fine_tune_callback: An optional function that takes as
                                   arguments the :class:`DBN` instance
                                   and the epoch, and is called for
                                   each epoch of fine tuning.

        :param random_state: An optional int used as the seed by the
                             random number generator.

        :param verbose: Debugging output.
        """

        if layer_sizes is None:
            layer_sizes = [-1, -1]

        if output_act_funct is None:
            output_act_funct = activationFunctions.Softmax()
        elif isinstance(output_act_funct, str):
            output_act_funct = getattr(activationFunctions, output_act_funct)()

        if random_state is not None:
            raise ValueError("random_sate must be an int")

        self.layer_sizes = layer_sizes
        self.scales = scales
        self.fan_outs = fan_outs
        self.output_act_funct = output_act_funct
        self.real_valued_vis = real_valued_vis
        self.use_re_lu = use_re_lu
        self.uniforms = uniforms

        self.learn_rates = learn_rates
        self.learn_rate_decays = learn_rate_decays
        self.learn_rate_minimums = learn_rate_minimums
        self.momentum = momentum
        self.l2_costs = l2_costs
        self.dropouts = dropouts
        self.nesterov = nesterov
        self.nest_compare = nest_compare
        self.rms_lims = rms_lims

        self.learn_rates_pretrain = learn_rates_pretrain
        self.momentum_pretrain = momentum_pretrain
        self.l2_costs_pretrain = l2_costs_pretrain
        self.nest_compare_pretrain = nest_compare_pretrain

        self.epochs = epochs
        self.epochs_pretrain = epochs_pretrain
        self.loss_funct = loss_funct
        self.use_dropout = True if dropouts else False
        self.minibatch_size = minibatch_size
        self.minibatches_per_epoch = minibatches_per_epoch

        self.pretrain_callback = pretrain_callback
        self.fine_tune_callback = fine_tune_callback
        self.random_state = random_state
        self.verbose = verbose

    def _fill_missing_layer_sizes(self, X, y):
        layer_sizes = self.layer_sizes
        if layer_sizes[0] == -1:  # n_feat
            layer_sizes[0] = X.shape[1]
        if layer_sizes[-1] == -1 and y is not None:  # n_classes
            layer_sizes[-1] = y.shape[1]

    def _vp(self, value):
        num_weights = len(self.layer_sizes) - 1
        if not hasattr(value, '__iter__'):
            value = [value] * num_weights
        return list(value)

    def _build_net(self, X, y=None):
        v = self._vp

        self._fill_missing_layer_sizes(X, y)
        if self.verbose:  # pragma: no cover
            print("[DBN] layers {}".format(self.layer_sizes))

        if self.random_state is not None:
            np.random.seed(self.random_state)

        net = buildDBN(
            self.layer_sizes,
            v(self.scales),
            v(self.fan_outs),
            self.output_act_funct,
            self.real_valued_vis,
            self.use_re_lu,
            v(self.uniforms),
            )

        return net

    def _configure_net_pretrain(self, net):
        v = self._vp

        self._configure_net_finetune(net)

        learn_rates = self.learn_rates_pretrain
        momentum = self.momentum_pretrain
        l2_costs = self.l2_costs_pretrain
        nest_compare = self.nest_compare_pretrain

        if learn_rates is None:
            learn_rates = self.learn_rates
        if momentum is None:
            momentum = self.momentum
        if l2_costs is None:
            l2_costs = self.l2_costs
        if nest_compare is None:
            nest_compare = self.nest_compare

        net.learnRates = v(learn_rates)
        net.momentum = momentum
        net.L2Costs = v(l2_costs)
        net.nestCompare = nest_compare

        return net

    def _configure_net_finetune(self, net):
        v = self._vp

        net.learnRates = v(self.learn_rates)
        net.momentum = self.momentum
        net.L2Costs = v(self.l2_costs)
        net.dropouts = v(self.dropouts)
        net.nesterov = self.nesterov
        net.nestCompare = self.nest_compare
        net.rmsLims = v(self.rms_lims)

        return net

    def _minibatches(self, X, y=None):
        while True:
            idx = np.random.randint(X.shape[0], size=(self.minibatch_size,))

            X_batch = X[idx]
            if hasattr(X_batch, 'todense'):
                X_batch = X_batch.todense()

            if y is not None:
                yield (X_batch, y[idx])
            else:
                yield X_batch

    def _onehot(self, y):
        return np.array(
            OneHotEncoder().fit_transform(y.reshape(-1, 1)).todense())

    def _num_mistakes(self, targets, outputs):
        if hasattr(targets, 'as_numpy_array'):  # pragma: no cover
            targets = targets.as_numpy_array()
        if hasattr(outputs, 'as_numpy_array'):
            outputs = outputs.as_numpy_array()
        return np.sum(outputs.argmax(1) != targets.argmax(1))

    def _learn_rate_adjust(self):
        if self.learn_rate_decays == 1.0:
            return

        learn_rate_decays = self._vp(self.learn_rate_decays)
        learn_rate_minimums = self._vp(self.learn_rate_minimums)

        for index, decay in enumerate(learn_rate_decays):
            new_learn_rate = self.net_.learnRates[index] * decay
            if new_learn_rate >= learn_rate_minimums[index]:
                self.net_.learnRates[index] = new_learn_rate

        if self.verbose >= 2:
            print("Learn rates: {}".format(self.net_.learnRates))

    def fit(self, X, y):
        if self.verbose:
            print("[DBN] fitting X.shape=%s" % (X.shape,))
        self._enc = LabelEncoder()
        y = self._enc.fit_transform(y)
        y = self._onehot(y)

        self.net_ = self._build_net(X, y)

        minibatches_per_epoch = self.minibatches_per_epoch
        if minibatches_per_epoch is None:
            minibatches_per_epoch = X.shape[0] / self.minibatch_size

        loss_funct = self.loss_funct
        if loss_funct is None:
            loss_funct = self._num_mistakes

        errors_pretrain = self.errors_pretrain_ = []
        losses_fine_tune = self.losses_fine_tune_ = []
        errors_fine_tune = self.errors_fine_tune_ = []

        if self.epochs_pretrain:
            self.epochs_pretrain = self._vp(self.epochs_pretrain)
            self._configure_net_pretrain(self.net_)
            for layer_index in range(len(self.layer_sizes) - 1):
                errors_pretrain.append([])
                if self.verbose:  # pragma: no cover
                    print("[DBN] Pre-train layer {}...".format(layer_index + 1))
                time0 = time()
                for epoch, err in enumerate(
                    self.net_.preTrainIth(
                        layer_index,
                        self._minibatches(X),
                        self.epochs_pretrain[layer_index],
                        minibatches_per_epoch,
                        )):
                    errors_pretrain[-1].append(err)
                    if self.verbose:  # pragma: no cover
                        print("  Epoch {}: err {}".format(epoch + 1, err))
                        elapsed = str(timedelta(seconds=time() - time0))
                        print("  ({})".format(elapsed.split('.')[0]))
                        time0 = time()
                    if self.pretrain_callback is not None:
                        self.pretrain_callback(
                            self, epoch + 1, layer_index)

        self._configure_net_finetune(self.net_)
        if self.verbose:  # pragma: no cover
            print("[DBN] Fine-tune...")
        time0 = time()
        for epoch, (loss, err) in enumerate(
            self.net_.fineTune(
                self._minibatches(X, y),
                self.epochs,
                minibatches_per_epoch,
                loss_funct,
                self.verbose,
                self.use_dropout,
                )):
            losses_fine_tune.append(loss)
            errors_fine_tune.append(err)
            self._learn_rate_adjust()
            if self.verbose:  # pragma: no cover
                print("Epoch {}:".format(epoch + 1))
                print("  loss {}".format(loss))
                print("  err  {}".format(err))
                elapsed = str(timedelta(seconds=time() - time0))
                print("  ({})".format(elapsed.split('.')[0]))
                time0 = time()
            if self.fine_tune_callback is not None:
                self.fine_tune_callback(self, epoch + 1)

    def predict(self, X):
        y_ind = np.argmax(self.predict_proba(X), axis=1)
        return self._enc.inverse_transform(y_ind)

    def predict_proba(self, X):
        if hasattr(X, 'todense'):
            return self._predict_proba_sparse(X)
        res = np.zeros((X.shape[0], self.layer_sizes[-1]))
        for i, el in enumerate(self.net_.predictions(X, asNumpy=True)):
            res[i] = el
        return res

    def _predict_proba_sparse(self, X):
        batch_size = self.minibatch_size
        res = []
        for i in range(0, X.shape[0], batch_size):
            X_batch = X[i:min(i + batch_size, X.shape[0])].todense()
            res.extend(self.net_.predictions(X_batch))
        return np.array(res).reshape(X.shape[0], -1)

    def score(self, X, y):
        loss_funct = self.loss_funct
        if loss_funct is None:
            loss_funct = self._num_mistakes

        outputs = self.predict_proba(X)
        targets = self._onehot(self._enc.transform(y))
        mistakes = loss_funct(outputs, targets)
        return - float(mistakes) / len(y) + 1

    @property
    def classes_(self):
        return self._enc.classes_



import sys

PY2 = sys.version_info[0] == 2

if PY2:
    from ConfigParser import ConfigParser
    from StringIO import StringIO
    import cPickle as pickle
    import __builtin__ as builtins

    basestring = basestring

    def chain_exception(exc1, exc2):
        exec("raise exc1, None, sys.exc_info()[2]")

else:
    from configparser import ConfigParser
    from io import StringIO
    import pickle as pickle
    import builtins

    basestring = str

    def chain_exception(exc1, exc2):
        exec("raise exc1 from exc2")






import numpy as np
import pytest
from sklearn.datasets import load_boston
from sklearn.datasets import fetch_mldata
from sklearn.preprocessing import StandardScaler
from sklearn.utils import shuffle

from lasagne.layers import Conv2DLayer
from lasagne.layers import DenseLayer
from lasagne.layers import InputLayer
from lasagne.layers import MaxPool2DLayer
from lasagne.layers import NonlinearityLayer
from lasagne.nonlinearities import softmax
from lasagne.updates import nesterov_momentum


@pytest.fixture(scope='session')
def NeuralNet():
    from nolearn.lasagne import NeuralNet
    return NeuralNet


@pytest.fixture
def nn(NeuralNet):
    return NeuralNet([('input', object())], input_shape=(10, 10))


@pytest.fixture(scope='session')
def mnist():
    dataset = fetch_mldata('mnist-original')
    X, y = dataset.data, dataset.target
    X = X.astype(np.float32) / 255.0
    y = y.astype(np.int32)
    return shuffle(X, y, random_state=42)


@pytest.fixture(scope='session')
def boston():
    dataset = load_boston()
    X, y = dataset.data, dataset.target
    # X, y = make_regression(n_samples=100000, n_features=13)
    X = StandardScaler().fit_transform(X).astype(np.float32)
    y = y.reshape(-1, 1).astype(np.float32)
    return shuffle(X, y, random_state=42)


class _OnEpochFinished:
    def __call__(self, nn, train_history):
        self.train_history = train_history
        if len(train_history) > 1:
            raise StopIteration()


@pytest.fixture(scope='session')
def X_train(mnist):
    X, y = mnist
    return X[:10000].reshape(-1, 1, 28, 28)


@pytest.fixture(scope='session')
def y_train(mnist):
    X, y = mnist
    return y[:10000]


@pytest.fixture(scope='session')
def X_test(mnist):
    X, y = mnist
    return X[60000:].reshape(-1, 1, 28, 28)


@pytest.fixture(scope='session')
def y_pred(net_fitted, X_test):
    return net_fitted.predict(X_test)


@pytest.fixture(scope='session')
def net(NeuralNet):
    l = InputLayer(shape=(None, 1, 28, 28))
    l = Conv2DLayer(l, name='conv1', filter_size=(5, 5), num_filters=8)
    l = MaxPool2DLayer(l, name='pool1', pool_size=(2, 2))
    l = Conv2DLayer(l, name='conv2', filter_size=(5, 5), num_filters=8)
    l = MaxPool2DLayer(l, name='pool2', pool_size=(2, 2))
    l = DenseLayer(l, name='hidden1', num_units=128)
    l = DenseLayer(l, name='output', nonlinearity=softmax, num_units=10)

    return NeuralNet(
        layers=l,

        update=nesterov_momentum,
        update_learning_rate=0.01,
        update_momentum=0.9,

        max_epochs=5,
        on_epoch_finished=[_OnEpochFinished()],
        verbose=99,
        )


@pytest.fixture(scope='session')
def net_fitted(net, X_train, y_train):
    return net.fit(X_train, y_train)


@pytest.fixture(scope='session')
def net_color_non_square(NeuralNet):
    l = InputLayer(shape=(None, 3, 20, 28))
    l = Conv2DLayer(l, name='conv1', filter_size=(5, 5), num_filters=1)
    l = MaxPool2DLayer(l, name='pool1', pool_size=(2, 2))
    l = Conv2DLayer(l, name='conv2', filter_size=(5, 5), num_filters=8)
    l = MaxPool2DLayer(l, name='pool2', pool_size=(2, 2))
    l = DenseLayer(l, name='hidden1', num_units=128)
    l = DenseLayer(l, name='output', nonlinearity=softmax, num_units=10)

    net = NeuralNet(
        layers=l,

        update=nesterov_momentum,
        update_learning_rate=0.01,
        update_momentum=0.9,

        max_epochs=1,
        )
    net.initialize()
    return net


@pytest.fixture(scope='session')
def net_with_nonlinearity_layer(NeuralNet):
    l = InputLayer(shape=(None, 1, 28, 28))
    l = Conv2DLayer(l, name='conv1', filter_size=(5, 5), num_filters=8)
    l = MaxPool2DLayer(l, name='pool1', pool_size=(2, 2))
    l = Conv2DLayer(l, name='conv2', filter_size=(5, 5), num_filters=8)
    l = MaxPool2DLayer(l, name='pool2', pool_size=(2, 2))
    l = DenseLayer(l, name='hidden1', num_units=128)
    l = DenseLayer(l, name='output', nonlinearity=softmax, num_units=10)
    l = NonlinearityLayer(l)

    net = NeuralNet(
        layers=l,

        update=nesterov_momentum,
        update_learning_rate=0.01,
        update_momentum=0.9,

        max_epochs=5,
        on_epoch_finished=[_OnEpochFinished()],
        verbose=99,
        )
    net.initialize()
    return net


@pytest.fixture
def net_no_conv(NeuralNet):
    l = InputLayer(shape=(None, 100))
    l = DenseLayer(l, name='output', nonlinearity=softmax, num_units=10)

    return NeuralNet(
        layers=l,

        update=nesterov_momentum,
        update_learning_rate=0.01,
        update_momentum=0.9,

        max_epochs=1,
        verbose=99,
        )



from functools import reduce
from operator import mul

from lasagne.layers import Layer
from lasagne.layers import Conv2DLayer
from lasagne.layers import MaxPool2DLayer
import numpy as np
from tabulate import tabulate

convlayers = [Conv2DLayer]
maxpoollayers = [MaxPool2DLayer]
try:
    from lasagne.layers.cuda_convnet import Conv2DCCLayer
    from lasagne.layers.cuda_convnet import MaxPool2DCCLayer
    convlayers.append(Conv2DCCLayer)
    maxpoollayers.append(MaxPool2DCCLayer)
except ImportError:
    pass
try:
    from lasagne.layers.dnn import Conv2DDNNLayer
    from lasagne.layers.dnn import MaxPool2DDNNLayer
    convlayers.append(Conv2DDNNLayer)
    maxpoollayers.append(MaxPool2DDNNLayer)
except ImportError:
    pass

try:
    from lasagne.layers.corrmm import Conv2DMMLayer
    convlayers.append(Conv2DMMLayer)
except ImportError:
    pass


class ansi:
    BLUE = '\033[94m'
    CYAN = '\033[36m'
    GREEN = '\033[32m'
    MAGENTA = '\033[35m'
    RED = '\033[31m'
    ENDC = '\033[0m'


def is_conv2d(layers):
    if isinstance(layers, Layer):
        return isinstance(layers, tuple(convlayers))
    return any([isinstance(layer, tuple(convlayers))
                for layer in layers])


def is_maxpool2d(layers):
    if isinstance(layers, Layer):
        return isinstance(layers, tuple(maxpoollayers))
    return any([isinstance(layer, tuple(maxpoollayers))
                for layer in layers])


def get_real_filter(layers, img_size):
    """Get the real filter sizes of each layer involved in
    convoluation. See Xudong Cao:
    https://www.kaggle.com/c/datasciencebowl/forums/t/13166/happy-lantern-festival-report-and-code
    This does not yet take into consideration feature pooling,
    padding, striding and similar gimmicks.
    """
    real_filter = np.zeros((len(layers), 2))
    conv_mode = True
    first_conv_layer = True
    expon = np.ones((1, 2))

    for i, layer in enumerate(layers[1:]):
        j = i + 1
        if not conv_mode:
            real_filter[j] = img_size
            continue

        if is_conv2d(layer):
            if not first_conv_layer:
                new_filter = np.array(layer.filter_size) * expon
                real_filter[j] = new_filter
            else:
                new_filter = np.array(layer.filter_size) * expon
                real_filter[j] = new_filter
                first_conv_layer = False
        elif is_maxpool2d(layer):
            real_filter[j] = real_filter[i]
            expon *= np.array(layer.pool_size)
        else:
            conv_mode = False
            real_filter[j] = img_size

    real_filter[0] = img_size
    return real_filter


def get_receptive_field(layers, img_size):
    """Get the real filter sizes of each layer involved in
    convoluation. See Xudong Cao:
    https://www.kaggle.com/c/datasciencebowl/forums/t/13166/happy-lantern-festival-report-and-code
    This does not yet take into consideration feature pooling,
    padding, striding and similar gimmicks.
    """
    receptive_field = np.zeros((len(layers), 2))
    conv_mode = True
    first_conv_layer = True
    expon = np.ones((1, 2))

    for i, layer in enumerate(layers[1:]):
        j = i + 1
        if not conv_mode:
            receptive_field[j] = img_size
            continue

        if is_conv2d(layer):
            if not first_conv_layer:
                last_field = receptive_field[i]
                new_field = (last_field + expon *
                             (np.array(layer.filter_size) - 1))
                receptive_field[j] = new_field
            else:
                receptive_field[j] = layer.filter_size
                first_conv_layer = False
        elif is_maxpool2d(layer):
            receptive_field[j] = receptive_field[i]
            expon *= np.array(layer.pool_size)
        else:
            conv_mode = False
            receptive_field[j] = img_size

    receptive_field[0] = img_size
    return receptive_field


def get_conv_infos(net, min_capacity=100. / 6, detailed=False):
    CYA = ansi.CYAN
    END = ansi.ENDC
    MAG = ansi.MAGENTA
    RED = ansi.RED

    layers = net.layers_.values()
    # assume that first layer is input layer
    img_size = layers[0].output_shape[2:]

    header = ['name', 'size', 'total', 'cap.Y', 'cap.X',
              'cov.Y', 'cov.X']
    if detailed:
        header += ['filter Y', 'filter X', 'field Y', 'field X']

    shapes = [layer.output_shape[1:] for layer in layers]
    totals = [str(reduce(mul, shape)) for shape in shapes]
    shapes = ['x'.join(map(str, shape)) for shape in shapes]
    shapes = np.array(shapes).reshape(-1, 1)
    totals = np.array(totals).reshape(-1, 1)

    real_filters = get_real_filter(layers, img_size)
    receptive_fields = get_receptive_field(layers, img_size)
    capacity = 100. * real_filters / receptive_fields
    capacity[np.logical_not(np.isfinite(capacity))] = 1
    img_coverage = 100. * receptive_fields / img_size
    layer_names = [layer.name if layer.name
                   else str(layer).rsplit('.')[-1].split(' ')[0]
                   for layer in layers]

    colored_names = []
    for name, (covy, covx), (capy, capx) in zip(
            layer_names, img_coverage, capacity):
        if (
                ((covy > 100) or (covx > 100)) and
                ((capy < min_capacity) or (capx < min_capacity))
        ):
            name = "{}{}{}".format(RED, name, END)
        elif (covy > 100) or (covx > 100):
            name = "{}{}{}".format(CYA, name, END)
        elif (capy < min_capacity) or (capx < min_capacity):
            name = "{}{}{}".format(MAG, name, END)
        colored_names.append(name)
    colored_names = np.array(colored_names).reshape(-1, 1)

    table = np.hstack((colored_names, shapes, totals, capacity, img_coverage))
    if detailed:
        table = np.hstack((table, real_filters.astype(int),
                           receptive_fields.astype(int)))

    return tabulate(table, header, floatfmt='.2f')



from collections import OrderedDict
import pickle

from lasagne.layers import Conv2DLayer
from lasagne.layers import DenseLayer
from lasagne.layers import MaxPool2DLayer
from lasagne.layers import InputLayer
from lasagne.nonlinearities import softmax
from lasagne.updates import nesterov_momentum
from mock import patch
from mock import Mock
import numpy
import pytest

from nolearn._compat import builtins


def test_print_log(mnist):
    from nolearn.lasagne import PrintLog

    nn = Mock(
        regression=False,
        custom_scores=[('my1', 0.99)],
        scores_train=[('my2', 0.98)],
        scores_valid=[('my3', 0.33)],
        )

    train_history = [{
        'epoch': 1,
        'train_loss': 0.8,
        'valid_loss': 0.7,
        'train_loss_best': False,
        'valid_loss_best': False,
        'valid_accuracy': 0.9,
        'my1': 0.99,
        'my2': 0.98,
        'my3': 0.33,
        'dur': 1.0,
        }]
    output = PrintLog().table(nn, train_history)
    assert output.split() == [
        'epoch',
        'trn',
        'loss',
        'val',
        'loss',
        'trn/val',
        'valid',
        'acc',
        'my2',
        'my3',
        'my1',
        'dur',
        '-------',
        '----------',
        '----------',
        '---------',
        '-----------',
        '-------',
        '-------',
        '-------',
        '-----',
        '1',
        '0.80000',
        '0.70000',
        '1.14286',
        '0.90000',
        '0.98000',
        '0.33000',
        '0.99000',
        '1.00s',
        ]


class TestSaveWeights():
    @pytest.fixture
    def SaveWeights(self):
        from nolearn.lasagne import SaveWeights
        return SaveWeights

    def test_every_n_epochs_true(self, SaveWeights):
        train_history = [{'epoch': 9, 'valid_loss': 1.1}]
        nn = Mock()
        handler = SaveWeights('mypath', every_n_epochs=3)
        handler(nn, train_history)
        assert nn.save_params_to.call_count == 1
        nn.save_params_to.assert_called_with('mypath')

    def test_every_n_epochs_false(self, SaveWeights):
        train_history = [{'epoch': 9, 'valid_loss': 1.1}]
        nn = Mock()
        handler = SaveWeights('mypath', every_n_epochs=4)
        handler(nn, train_history)
        assert nn.save_params_to.call_count == 0

    def test_only_best_true_single_entry(self, SaveWeights):
        train_history = [{'epoch': 9, 'valid_loss': 1.1}]
        nn = Mock()
        handler = SaveWeights('mypath', only_best=True)
        handler(nn, train_history)
        assert nn.save_params_to.call_count == 1

    def test_only_best_true_two_entries(self, SaveWeights):
        train_history = [
            {'epoch': 9, 'valid_loss': 1.2},
            {'epoch': 10, 'valid_loss': 1.1},
            ]
        nn = Mock()
        handler = SaveWeights('mypath', only_best=True)
        handler(nn, train_history)
        assert nn.save_params_to.call_count == 1

    def test_only_best_false_two_entries(self, SaveWeights):
        train_history = [
            {'epoch': 9, 'valid_loss': 1.2},
            {'epoch': 10, 'valid_loss': 1.3},
            ]
        nn = Mock()
        handler = SaveWeights('mypath', only_best=True)
        handler(nn, train_history)
        assert nn.save_params_to.call_count == 0

    def test_with_path_interpolation(self, SaveWeights):
        train_history = [{'epoch': 9, 'valid_loss': 1.1}]
        nn = Mock()
        handler = SaveWeights('mypath-{epoch}-{timestamp}-{loss}.pkl')
        handler(nn, train_history)
        path = nn.save_params_to.call_args[0][0]
        assert path.startswith('mypath-0009-2')
        assert path.endswith('-1.1.pkl')

    def test_pickle(self, SaveWeights):
        train_history = [{'epoch': 9, 'valid_loss': 1.1}]
        nn = Mock()
        with patch('nolearn.lasagne.handlers.pickle') as pickle:
            with patch.object(builtins, 'open') as mock_open:
                handler = SaveWeights('mypath', every_n_epochs=3, pickle=True)
                handler(nn, train_history)

        mock_open.assert_called_with('mypath', 'wb')
        pickle.dump.assert_called_with(nn, mock_open().__enter__(), -1)


class TestRememberBestWeights:
    @pytest.fixture
    def RememberBestWeights(self):
        from nolearn.lasagne.handlers import RememberBestWeights
        return RememberBestWeights

    @pytest.mark.parametrize('loss_name', ['valid_loss', 'my_loss'])
    def test_simple(self, RememberBestWeights, loss_name):
        nn1, nn2, nn3 = Mock(), Mock(), Mock()
        rbw = RememberBestWeights(loss=loss_name)
        train_history = []

        train_history.append({'epoch': 1, loss_name: 1.0})
        rbw(nn1, train_history)
        assert rbw.best_weights is nn1.get_all_params_values()

        train_history.append({'epoch': 2, loss_name: 1.1})
        rbw(nn2, train_history)
        assert rbw.best_weights is nn1.get_all_params_values()

        train_history.append({'epoch': 3, loss_name: 0.9})
        rbw(nn3, train_history)
        assert rbw.best_weights is nn3.get_all_params_values()

    def test_custom_score(self, RememberBestWeights):
        nn1, nn2, nn3 = Mock(), Mock(), Mock()
        rbw = RememberBestWeights(score='myscr')
        train_history = []

        train_history.append({'epoch': 1, 'myscr': 1.0})
        rbw(nn1, train_history)
        assert rbw.best_weights is nn1.get_all_params_values()

        train_history.append({'epoch': 2, 'myscr': 1.1})
        rbw(nn2, train_history)
        assert rbw.best_weights is nn2.get_all_params_values()

        train_history.append({'epoch': 3, 'myscr': 0.9})
        rbw(nn3, train_history)
        assert rbw.best_weights is nn2.get_all_params_values()

    def test_restore(self, RememberBestWeights):
        nn = Mock()
        rbw = RememberBestWeights()
        train_history = []
        train_history.append({'epoch': 1, 'valid_loss': 1.0})
        rbw(nn, train_history)
        rbw.restore(nn, train_history)
        nn.load_params_from.assert_called_with(nn.get_all_params_values())
        nn.load_params_from.assert_called_with(rbw.best_weights)


class TestPrintLayerInfo():
    @pytest.fixture(scope='session')
    def X_train(self, mnist):
        X, y = mnist
        return X[:100].reshape(-1, 1, 28, 28)

    @pytest.fixture(scope='session')
    def y_train(self, mnist):
        X, y = mnist
        return y[:100]

    @pytest.fixture(scope='session')
    def nn(self, NeuralNet, X_train, y_train):
        nn = NeuralNet(
            layers=[
                ('input', InputLayer),
                ('dense0', DenseLayer),
                ('dense1', DenseLayer),
                ('output', DenseLayer),
                ],
            input_shape=(None, 1, 28, 28),
            output_num_units=10,
            output_nonlinearity=softmax,

            more_params=dict(
                dense0_num_units=16,
                dense1_num_units=16,
                ),

            update=nesterov_momentum,
            update_learning_rate=0.01,
            update_momentum=0.9,

            max_epochs=3,
            )
        nn.initialize()

        return nn

    @pytest.fixture(scope='session')
    def cnn(self, NeuralNet, X_train, y_train):
        nn = NeuralNet(
            layers=[
                ('input', InputLayer),
                ('conv1', Conv2DLayer),
                ('conv2', Conv2DLayer),
                ('pool2', MaxPool2DLayer),
                ('conv3', Conv2DLayer),
                ('output', DenseLayer),
                ],
            input_shape=(None, 1, 28, 28),
            output_num_units=10,
            output_nonlinearity=softmax,

            more_params=dict(
                conv1_filter_size=5, conv1_num_filters=16,
                conv2_filter_size=3, conv2_num_filters=16,
                pool2_pool_size=8, pool2_ignore_border=False,
                conv3_filter_size=3, conv3_num_filters=16,
                hidden1_num_units=16,
                ),

            update=nesterov_momentum,
            update_learning_rate=0.01,
            update_momentum=0.9,

            max_epochs=3,
            )

        nn.initialize()
        return nn

    @pytest.fixture
    def is_conv2d(self):
        from nolearn.lasagne.util import is_conv2d
        return is_conv2d

    @pytest.fixture
    def is_maxpool2d(self):
        from nolearn.lasagne.util import is_maxpool2d
        return is_maxpool2d

    @pytest.fixture
    def print_info(self):
        from nolearn.lasagne.handlers import PrintLayerInfo
        return PrintLayerInfo()

    def test_is_conv2d_net_false(self, nn, is_conv2d):
        assert is_conv2d(nn.layers_.values()) is False

    def test_is_conv2d_net_true(self, cnn, is_conv2d):
        assert is_conv2d(cnn.layers_.values()) is True

    def test_is_conv2d_layer(self, nn, cnn, is_conv2d):
        assert is_conv2d(nn.layers_['input']) is False
        assert is_conv2d(cnn.layers_['pool2']) is False
        assert is_conv2d(cnn.layers_['conv1']) is True

    def test_is_maxpool2d_net_false(self, nn, is_maxpool2d):
        assert is_maxpool2d(nn.layers_.values()) is False

    def test_is_maxpool2d_net_true(self, cnn, is_maxpool2d):
        assert is_maxpool2d(cnn.layers_.values()) is True

    def test_is_maxpool2d_layer(self, nn, cnn, is_maxpool2d):
        assert is_maxpool2d(nn.layers_['input']) is False
        assert is_maxpool2d(cnn.layers_['pool2']) is True
        assert is_maxpool2d(cnn.layers_['conv1']) is False

    def test_print_layer_info_greeting(self, nn, print_info):
        # number of learnable parameters is weights + biases:
        # 28 * 28 * 16 + 16 + 16 * 16 + 16 + 16 * 10 + 10 = 13002
        expected = '# Neural Network with 13002 learnable parameters\n'
        message = print_info._get_greeting(nn)
        assert message == expected

    def test_print_layer_info_plain_nn(self, nn, print_info):
        expected = """\
  #  name    size
---  ------  -------
  0  input   1x28x28
  1  dense0  16
  2  dense1  16
  3  output  10"""
        message = print_info._get_layer_info_plain(nn)
        assert message == expected

    def test_print_layer_info_plain_cnn(self, cnn, print_info):
        expected = """\
  #  name    size
---  ------  --------
  0  input   1x28x28
  1  conv1   16x24x24
  2  conv2   16x22x22
  3  pool2   16x3x3
  4  conv3   16x1x1
  5  output  10"""
        message = print_info._get_layer_info_plain(cnn)
        assert message == expected

    def test_print_layer_info_conv_cnn(self, cnn, print_info):
        expected = """\
name    size        total    cap.Y    cap.X    cov.Y    cov.X
------  --------  -------  -------  -------  -------  -------
input   1x28x28       784   100.00   100.00   100.00   100.00
conv1   16x24x24     9216   100.00   100.00    17.86    17.86
conv2   16x22x22     7744    42.86    42.86    25.00    25.00
pool2   16x3x3        144    42.86    42.86    25.00    25.00
conv3   16x1x1         16   104.35   104.35    82.14    82.14
output  10             10   100.00   100.00   100.00   100.00"""
        message, legend = print_info._get_layer_info_conv(cnn)
        assert message == expected

        expected = """
Explanation
    X, Y:    image dimensions
    cap.:    learning capacity
    cov.:    coverage of image
    \x1b[35mmagenta\x1b[0m: capacity too low (<1/6)
    \x1b[36mcyan\x1b[0m:    image coverage too high (>100%)
    \x1b[31mred\x1b[0m:     capacity too low and coverage too high
"""
        assert legend == expected


class TestWeightLog:
    @pytest.fixture
    def WeightLog(self):
        from nolearn.lasagne import WeightLog
        return WeightLog

    @pytest.fixture
    def nn(self):
        nn = Mock()
        nn.get_all_params_values.side_effect = [
            OrderedDict([
                ('layer1', numpy.array([[-1, -2]])),
                ('layer2', numpy.array([[3, 4]])),
                ]),
            OrderedDict([
                ('layer1', numpy.array([[-2, -3]])),
                ('layer2', numpy.array([[5, 7]])),
                ]),
            ]
        return nn

    def test_history(self, WeightLog, nn):
        wl = WeightLog()
        wl(nn, None)
        wl(nn, None)

        assert wl.history[0] == {
            'layer1_0 wdiff': 0,
            'layer1_0 wabsmean': 1.5,
            'layer1_0 wmean': -1.5,

            'layer2_0 wdiff': 0,
            'layer2_0 wabsmean': 3.5,
            'layer2_0 wmean': 3.5,
            }

        assert wl.history[1]['layer1_0 wdiff'] == 1.0
        assert wl.history[1]['layer2_0 wdiff'] == 2.5

    def test_save_to(self, WeightLog, nn, tmpdir):
        save_to = tmpdir.join("hello.csv")
        wl = WeightLog(save_to=save_to.strpath, write_every=1)
        wl(nn, None)
        wl(nn, None)

        assert save_to.readlines() == [
            'layer1_0 wdiff,layer1_0 wabsmean,layer1_0 wmean,'
            'layer2_0 wdiff,layer2_0 wabsmean,layer2_0 wmean\n',
            '0.0,1.5,-1.5,0.0,3.5,3.5\n',
            '1.0,2.5,-2.5,2.5,6.0,6.0\n',
            ]

    def test_pickle(self, WeightLog, nn, tmpdir):
        save_to = tmpdir.join("hello.csv")
        pkl = tmpdir.join("hello.pkl")
        wl = WeightLog(save_to=save_to.strpath, write_every=1)
        wl(nn, None)

        with open(pkl.strpath, 'wb') as f:
            pickle.dump(wl, f)

        with open(pkl.strpath, 'rb') as f:
            wl = pickle.load(f)

        wl(nn, None)
        assert save_to.readlines() == [
            'layer1_0 wdiff,layer1_0 wabsmean,layer1_0 wmean,'
            'layer2_0 wdiff,layer2_0 wabsmean,layer2_0 wmean\n',
            '0.0,1.5,-1.5,0.0,3.5,3.5\n',
            '1.0,2.5,-2.5,2.5,6.0,6.0\n',
            ]



import os

from setuptools import setup, find_packages

version = '0.6.0'

here = os.path.abspath(os.path.dirname(__file__))
try:
    README = open(os.path.join(here, 'README.rst')).read()
    CHANGES = open(os.path.join(here, 'CHANGES.txt')).read()
except IOError:
    README = CHANGES = ''

install_requires = [
    'gdbn',
    'joblib',
    'scikit-learn',
    'tabulate',
    'Lasagne',
    'Theano',
    ]

visualization_require = [
    'matplotlib',
    'pydotplus',
    'ipython'
    ]

tests_require = [
    'mock',
    'pytest',
    'pytest-cov',
    'pytest-flakes',
    'pytest-pep8',
    ]

docs_require = [
    'Sphinx',
    ]

all_require = (visualization_require + tests_require + docs_require)

setup(name='nolearn',
      version=version,
      description="scikit-learn compatible neural network library",
      long_description='\n\n'.join([README, CHANGES]),
      classifiers=[
          "Development Status :: 4 - Beta",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.4",
          ],
      keywords='',
      author='Daniel Nouri',
      author_email='daniel.nouri@gmail.com',
      url='https://github.com/dnouri/nolearn',
      license='MIT',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      install_requires=install_requires,
      extras_require={
          'visualization': visualization_require,
          'testing': tests_require,
          'docs': docs_require,
          'all': all_require,
          },
      )



from collections import OrderedDict
from csv import DictWriter
from datetime import datetime
from functools import reduce
import operator
import sys

import numpy
from tabulate import tabulate

from .._compat import pickle
from .util import ansi
from .util import get_conv_infos
from .util import is_conv2d


class PrintLog:
    def __init__(self):
        self.first_iteration = True

    def __call__(self, nn, train_history):
        print(self.table(nn, train_history))
        sys.stdout.flush()

    def table(self, nn, train_history):
        info = train_history[-1]

        info_tabulate = OrderedDict([
            ('epoch', info['epoch']),
            ('trn loss', "{}{:.5f}{}".format(
                ansi.CYAN if info['train_loss_best'] else "",
                info['train_loss'],
                ansi.ENDC if info['train_loss_best'] else "",
                )),
            ('val loss', "{}{:.5f}{}".format(
                ansi.GREEN if info['valid_loss_best'] else "",
                info['valid_loss'],
                ansi.ENDC if info['valid_loss_best'] else "",
                )),
            ('trn/val', info['train_loss'] / info['valid_loss']),
            ])

        if not nn.regression:
            info_tabulate['valid acc'] = info['valid_accuracy']

        for name, func in nn.scores_train:
            info_tabulate[name] = info[name]

        for name, func in nn.scores_valid:
            info_tabulate[name] = info[name]

        if nn.custom_scores:
            for custom_score in nn.custom_scores:
                info_tabulate[custom_score[0]] = info[custom_score[0]]

        info_tabulate['dur'] = "{:.2f}s".format(info['dur'])

        tabulated = tabulate(
            [info_tabulate], headers="keys", floatfmt='.5f')

        out = ""
        if self.first_iteration:
            out = "\n".join(tabulated.split('\n', 2)[:2])
            out += "\n"
            self.first_iteration = False

        out += tabulated.rsplit('\n', 1)[-1]
        return out


class SaveWeights:
    def __init__(self, path, every_n_epochs=1, only_best=False,
                 pickle=False, verbose=0):
        self.path = path
        self.every_n_epochs = every_n_epochs
        self.only_best = only_best
        self.pickle = pickle
        self.verbose = verbose

    def __call__(self, nn, train_history):
        if self.only_best:
            this_loss = train_history[-1]['valid_loss']
            best_loss = min([h['valid_loss'] for h in train_history])
            if this_loss > best_loss:
                return

        if train_history[-1]['epoch'] % self.every_n_epochs != 0:
            return

        format_args = {
            'loss': train_history[-1]['valid_loss'],
            'timestamp': datetime.now().strftime('%Y-%m-%d-%H-%M-%S'),
            'epoch': '{:04d}'.format(train_history[-1]['epoch']),
            }
        path = self.path.format(**format_args)

        if self.verbose:
            print("Writing {}".format(path))

        if self.pickle:
            with open(path, 'wb') as f:
                pickle.dump(nn, f, -1)
        else:
            nn.save_params_to(path)


class _RestoreBestWeights:
    def __init__(self, remember):
        self.remember = remember

    def __call__(self, nn, train_history):
        nn.load_params_from(self.remember.best_weights)
        if self.remember.verbose:
            print("Loaded best weights from epoch {} where {} was {}".format(
                self.remember.best_weights_epoch,
                self.remember.score or self.remember.loss,
                self.remember.best_weights_loss * (
                    -1 if self.remember.score else 1),
                ))


class RememberBestWeights:
    def __init__(self, loss='valid_loss', score=None, verbose=1):
        self.loss = loss
        self.score = score
        self.verbose = 1
        self.best_weights = None
        self.best_weights_loss = sys.maxsize
        self.best_weights_epoch = None
        self.restore = _RestoreBestWeights(self)

    def __call__(self, nn, train_history):
        key = self.score if self.score is not None else self.loss

        curr_loss = train_history[-1][key]
        if self.score:
            curr_loss *= -1

        if curr_loss < self.best_weights_loss:
            self.best_weights = nn.get_all_params_values()
            self.best_weights_loss = curr_loss
            self.best_weights_epoch = train_history[-1]['epoch']


class PrintLayerInfo:
    def __init__(self):
        pass

    def __call__(self, nn, train_history=None):
        if train_history:
            return

        message = self._get_greeting(nn)
        print(message)
        print("## Layer information")
        print("")

        layers_contain_conv2d = is_conv2d(nn.layers_.values())
        if not layers_contain_conv2d or (nn.verbose < 2):
            layer_info = self._get_layer_info_plain(nn)
            legend = None
        else:
            layer_info, legend = self._get_layer_info_conv(nn)
        print(layer_info)
        if legend is not None:
            print(legend)
        print("")
        sys.stdout.flush()

    @staticmethod
    def _get_greeting(nn):
        shapes = [param.get_value().shape for param in
                  nn.get_all_params(trainable=True) if param]
        nparams = reduce(operator.add, [reduce(operator.mul, shape) for
                                        shape in shapes])
        message = ("# Neural Network with {} learnable parameters"
                   "\n".format(nparams))
        return message

    @staticmethod
    def _get_layer_info_plain(nn):
        nums = list(range(len(nn.layers_)))
        names = [layer.name for layer in nn.layers_.values()]
        output_shapes = ['x'.join(map(str, layer.output_shape[1:]))
                         for layer in nn.layers_.values()]

        table = OrderedDict([
            ('#', nums),
            ('name', names),
            ('size', output_shapes),
        ])
        layer_infos = tabulate(table, 'keys')
        return layer_infos

    @staticmethod
    def _get_layer_info_conv(nn):
        if nn.verbose > 2:
            detailed = True
        else:
            detailed = False

        layer_infos = get_conv_infos(nn, detailed=detailed)

        mag = "{}{}{}".format(ansi.MAGENTA, "magenta", ansi.ENDC)
        cya = "{}{}{}".format(ansi.CYAN, "cyan", ansi.ENDC)
        red = "{}{}{}".format(ansi.RED, "red", ansi.ENDC)
        legend = (
            "\nExplanation"
            "\n    X, Y:    image dimensions"
            "\n    cap.:    learning capacity"
            "\n    cov.:    coverage of image"
            "\n    {}: capacity too low (<1/6)"
            "\n    {}:    image coverage too high (>100%)"
            "\n    {}:     capacity too low and coverage too high\n"
            "".format(mag, cya, red)
            )

        return layer_infos, legend


class WeightLog:
    """Keep a log of your network's weights and weight changes.

    Pass instances of :class:`WeightLog` as an `on_batch_finished`
    handler into your network.
    """
    def __init__(self, save_to=None, write_every=8):
        """
        :param save_to: If given, `save_to` must be a path into which
                        I will write weight statistics in CSV format.
        """
        self.last_weights = None
        self.history = []
        self.save_to = save_to
        self.write_every = write_every
        self._dictwriter = None
        self._save_to_file = None

    def __call__(self, nn, train_history):
        weights = nn.get_all_params_values()

        if self.save_to and self._dictwriter is None:
            fieldnames = []
            for key in weights.keys():
                for i, p in enumerate(weights[key]):
                    fieldnames.extend([
                        '{}_{} wdiff'.format(key, i),
                        '{}_{} wabsmean'.format(key, i),
                        '{}_{} wmean'.format(key, i),
                        ])

            newfile = self.last_weights is None
            if newfile:
                self._save_to_file = open(self.save_to, 'w')
            else:
                self._save_to_file = open(self.save_to, 'a')
            self._dictwriter = DictWriter(self._save_to_file, fieldnames)
            if newfile:
                self._dictwriter.writeheader()

        entry = {}
        lw = self.last_weights if self.last_weights is not None else weights
        for key in weights.keys():
            for i, (p1, p2) in enumerate(zip(lw[key], weights[key])):
                wdiff = numpy.abs(p1 - p2).mean()
                wabsmean = numpy.abs(p2).mean()
                wmean = p2.mean()
                entry.update({
                    '{}_{} wdiff'.format(key, i): wdiff,
                    '{}_{} wabsmean'.format(key, i): wabsmean,
                    '{}_{} wmean'.format(key, i): wmean,
                    })
        self.history.append(entry)

        if self.save_to:
            if len(self.history) % self.write_every == 0:
                self._dictwriter.writerows(self.history[-self.write_every:])
                self._save_to_file.flush()

        self.last_weights = weights

    def __getstate__(self):
        state = dict(self.__dict__)
        state['_save_to_file'] = None
        state['_dictwriter'] = None
        return state
